# -*- coding:utf-8 -*-
"""
This is main class that is used for whole processing logic for sklearn.

@author: Guangqiang.lu
"""
from sklearn.base import TransformerMixin
from auto_ml.utils.paths import get_file_base_name


class Process(TransformerMixin):
    """
    This is whole class that is used for pre-processing logic, just give a direction.
    Here for init function that we want to get just class name for later process.
    """
    def __init__(self):
        self.name = self.__class__.__name__
        self.estimator = None

    def fit(self, x, y=None):
        """
        For whole processing logic should provide with label, so that even we don't use
        it, we could just follow sklearn logic.
        :param data:
        :param label:
        :return:
        """
        self.estimator.fit(x, y=y)

    def fit_transform(self, data, y=None):
        """
        Make parent logic just like sklearn.
        :param data:
        :return:
        """
        self.fit(data, y=y)
        return self.transform(data, y=y)

    def transform(self, data, y=None):
        return self.estimator.transform(data)


if __name__ == '__main__':
    p = Process()
    print(p.name)



# -*- coding:utf-8 -*-
"""
Use a class get generate the hyper-parameters

@author: Guangqiang.lu
"""
import numpy as np
from collections import defaultdict


class ConfigSpace(object):
    def __init__(self):
        self.hyper_space = defaultdict(list)

    def add_hyper(self, hyper_list):
        if not isinstance(hyper_list, list):
            hyper_list = [hyper_list]

        for hyper in hyper_list:
            hyper_name = hyper.name
            for value in hyper.get_values():
                self.hyper_space[hyper_name].append(value)
        return self

    def get_hypers(self):
        # for key, values in self.hyper_space.items():
        #     print("key: ", key)
        #     print("Values: ", values)
            # print("Get hyper key: {}, {}".format(key, '\t'.join(str(values))))
        res_dict = dict()
        for k, v in self.hyper_space.items():
            res_dict[k] = v

        return res_dict


class Hyperarameter(object):
    """
    Whole hyperparameter parent class that use random logic
    """
    def __init__(self, name):
        self.name = name

    def get_values(self):
        raise NotImplementedError


class NormalHyperameter(Hyperarameter):
    def __init__(self, name, loc=0, scale=1, size=5):
        super().__init__(name=name)
        self.name = name
        self.loc = loc if loc > 0 else 0
        self.scale = scale if scale > 0 else 1
        self.size = size

    def get_values(self):
        """
        get sorted sampled values list
        :return: sorted list
        """
        samples = np.random.normal(loc=self.loc, scale=self.scale, size=self.size).tolist()
        return sorted(samples)


class UniformHyperparameter(Hyperarameter):
    def __init__(self, name, low=0, high=100, size=5):
        super().__init__(name=name)
        self.name = name
        self.low = low if low > 0 else 0
        self.high = high if high > 0 else 1
        self.size = size

    def get_values(self):
        """
        get uniform samples with uniform distribution
        :return:
        """
        samples = np.random.uniform(low=self.low, high=self.high, size=self.size).tolist()
        return sorted(samples)


class CategoryHyperparameter(Hyperarameter):
    def __init__(self, name, categories=[], default=None):
        super().__init__(name=name)
        self.name = name
        self.categories = categories
        self.default = default

    def get_values(self):
        """
        if and only if there are too many choices that need to
        choose, otherwise just return the whole categories
        if default is provided, then just use the default value.
        :return:
        """
        if self.default is not None:
            samples = [self.default]
        else:
            samples = self.categories

        return samples


class GridHyperparameter(Hyperarameter):
    """
    Use grid search to find whole wanted values
    """
    def __init__(self, name, values):
        self.name = name
        self.values = values

    def get_values(self):
        return self.values




# -*- coding:utf-8 -*-
"""
Whole utils that could be used for data convention transformation.

author: Guangqiang.lu
"""
import numpy as np
import hashlib
import pandas as pd
import scipy.sparse as sp
from sklearn.utils.validation import check_X_y, check_array
from sklearn.utils.multiclass import type_of_target

from auto_ml.utils.CONSTANT import *
from auto_ml.metrics.scorer import *


def ensure_data_without_nan(data):
    """
    this is used to convert data to ensure that
    there isn't anything nan or inf value.
    replace with inf and nan value with mean value for continuous
    data, with most frequent value for categorical feature.
    with each column.
    :param data: array type
    :return: array data
    """
    # this is to use DataFrame to make do the missing value filling
    if len(data.shape) == 1:
        data = data.reshape(-1, 1)
    # convert inf to nan value
    data[~np.isfinite(data)] = np.nan

    df = pd.DataFrame(data, columns=range(data.shape[1]))
    missing_cols = list(df.isnull().sum().index)
    df[missing_cols].fillna(df[missing_cols].mean())

    return df.values


def nan_col_index(data):
    """
    this is to get the nan columns index that contain inf value
    :param data: array type
    :return: index with True if nan or inf value
    """
    nan_index = np.any(~np.isfinite(data), axis=0)
    return nan_index


def is_sparse(data):
    """
    this is to check whether the data is sparse or not.
    :param data: array type
    :return: True is sparse data else False
    """
    if sp.issparse(data):
        return True
    else:
        return False


def check_data_and_label(data, label):
    """
    this is to check with data and label, we should
    ensure the data and label that we need.
    Also we support with sparse data.
    :param data: array-like
    :param label: array-like
    :return: checked data and label
    """
    check_label(label)
    # Here is to ensure that the data is 2D
    data = ensure_2d_data(data)

    # For `check_X_y` is only used for the numerical data type. So data validation shouldn't happen here.
    # Should after the processing logic finished.
    # data, label = check_X_y(data, label, accept_sparse=True)

    return data, label


def ensure_2d_data(data):
    """
    this is to ensure without nan or inf value,
    as for training step, we don't support nan value,
    this function should be used check data step later.
    :param data: array-like
    :return: array or raise error
    """
    if len(data.shape) == 1:
        # we have to ensure data is 2D
        data = data.reshape(-1, 1)

    # for `check_array` is only used for `float` type. Here we just want to ensure the data is 2D
    # return check_array(data)
    return data


def check_label(y):
    """
    this is to ensure we have satisfied target label,
    as we only support with CONSTANT type of target.
    :param y: array like
    :return:
        Type of the problem that are supported with a number based on the CONSTANT module.
        raise error not supported
    """
    # we don't support with target value with nan value
    if nan_col_index(y):
        raise ValueError("We don't support with nan-label! Please check it!")

    type_y = type_of_target(y)
    # current support target type
    supported_type = TASK_TO_STRING.values()
    if type_y not in supported_type:
        raise NotImplementedError("Don't support: %s type of problem" % type_y)

    type_y = STRING_TO_TASK.get(type_y)

    return type_y


def hash_dataset_name(data):
    """
    this is to hash the data to get the name of dataset
    :param data: array-like
    :return: string of a dataset as a name
    """
    md5 = hashlib.md5()
    if sp.issparse(data):
        md5.update(data.data)
        md5.update(str(data.shape).encode('utf-8'))
    else:
        md5.update(data)
    dataset_name = md5.hexdigest()

    return dataset_name


def convert_null_to_none(value):
    """
    this is to convert '' or nan to None for a list
    :param value: list or value
    :return: value
    """
    if isinstance(value, list):
        value = [None if x == '' else x for x in value]
    else:
        value = None

    return value


def is_categorical_type(series):
    """
    To check the data type of the series
    :param series:
    :return:
    """
    # after convert dataframe into array, if there is anything string, then others will be string too.
    # so we couldn't fit with array data type that has string type....
    from pandas.api.types import (is_bool_dtype, is_categorical_dtype, is_object_dtype, is_numeric_dtype,
                                  is_string_dtype, is_datetime64_dtype, is_timedelta64_dtype, is_integer_dtype)

    return is_string_dtype(series) or is_categorical_dtype(series) or \
           is_object_dtype(series) or is_bool_dtype(series) or is_integer_dtype(series)


def get_type_problem(y):
    """
    To check what type of the label dataset.
    :param y:
    :return:
    """
    label_type = check_label(y)

    if label_type in CLASSIFICTION_TASK:
        return 'classification'
    elif label_type in REGRESSION_TASK:
        return 'regression'
    else:
        raise ValueError("When to check label type based label data, "
                         "get not supported type: {}".format(label_type))


def get_scorer_based_on_target(y):
    """
    Based on the target to return the type of the problem.
    :param y:
    :return:
    """
    # for different problem use different scorer
    # self.type_of_problem = self._get_type_problem(y)
    type_of_problem = get_type_problem(y)

    if type_of_problem == 'classification':
        scorer = accuracy
    elif type_of_problem == 'regression':
        scorer = mean_squared_error
    else:
        raise ValueError("When to score data get not "
                         "supported type of problem: {}".format(type_of_problem))

    return scorer


def get_num_classes_based_on_label(label):
    if len(label.shape) > 2:
        if label.shape[1] == 1:
            return len(np.unique(label[:, 0]))
        else:
            # currently is with multi-classes only, for multi-label should change this 
            return label.shape[1]
    else:
        return len(np.unique(label))


if __name__ == '__main__':
    from sklearn.datasets import load_iris
    x, y = load_iris(return_X_y=True)
    from auto_ml.test.get_test_data import get_training_data
    x, y = get_training_data()

    print(hash_dataset_name(x))
    # print(ensure_no_nan_value(x))
    # print(check_data_and_label(x, y))
    print(get_num_classes_based_on_label(y))


# -*- coding:utf-8 -*-
"""
This class is the main class that whole sklearn
estimators should inherit from this class, and
this class will set with the training step information,
like which folder to save the trained model, how much
time that total model training step, each algorithm
could be fitted time and how big the data set or something
useful in this class.
author: Guangqiang.lu
"""
from sklearn.base import BaseEstimator
from auto_ml.automl import ClassificationAutoML


class Estimators(BaseEstimator):
    """
    based on sklearn base estimator that we could fix
    whole estimators.
    """
    def __init__(self, total_times=3600,
                 each_task_time=300,
                 ensemble_size=5,
                 ensemble_nbest_size=10,
                 data_memory_limit=1024,
                 include_estimators=None,
                 include_processors=None,
                 exclude_estimators=None,
                 exclude_processors=None,
                 models_folder=None,
                 tmp_models_folder=None,
                 delete_tmp_models_folder_after_finish=True,
                 delete_models_folder_after_finish=False,
                 n_jobs=1,
                 ):
        """
        This is main class for whole sklearn processor class.

        :param total_times: int type, default=3600
            how much time that could be used to train models.
        :param each_task_time: int type, default=300
            how much time for each algorithm that could be trained.
        :param ensemble_size: int type, default=5
            how many ensembles models could be used to make the ensemble
            models, if make this bigger, the more chance to find a better
            model.
        :param ensemble_nbest_size: int type, default=10
            how many already trained model could be used to make the final
            ensemble model.
        :param data_memory_limit: int type, default=1024
            how big data could be used for training models.
        :param include_estimators: list(object), default=None
            which algorithms could just be used to fit the model, if default,
            then whole models will be fitted, if with these algorithms chosen,
            then just use these algorithms.
        :param include_processors: list(object), default=None
            which processing algorithms should be used to process the data.
            default will use whole predefined algorithms to process the data,
            otherwise will just use the chosen algorithms.
        :param exclude_estimators: list(object), default=None
            which algorithms should be excluded from the training algorithms,
            if these algorithms are chosen, then just remove these algorithms
            that could be used to fit the model.
        :param exclude_processors: list(object), default=None
            which processing algorithms could be excluded for the data, in case
            that we don't want to use to process the data.
        :param models_folder: str type, default=None
            the local folder the should be used to store the final combined model
            that could be used to put it into production.
        :param tmp_models_folder: str type, default=None
            which temperate folder could be used for whole training step to store
            the already trained model.
        :param delete_tmp_models_folder_after_finish: Boolean, default=True
            whether or not to delete the temperate folder for the models folder,
            default is true to delete the whole folders for stepping models.
        :param delete_models_folder_after_finish: Boolean, default=False
            whether or not to delete the already trained final ensemble or selected
            models that are stored in that folder, default is not to delete the folder.
        :param n_jobs: int type, default=None
            how many processors should be used for training step, default is None means
            to use whole processors, if negative then will also use whole processors.
        """
        self.total_times = total_times
        self.each_task_time = each_task_time
        self.ensemble_size = ensemble_size
        self.ensemble_nbest_size = ensemble_nbest_size
        self.data_memory_limit = data_memory_limit
        self.include_estimators = include_estimators
        self.include_processors = include_processors
        self.exclude_estimators = exclude_estimators
        self.exclude_processors = exclude_processors
        self.models_folder = models_folder
        self.tmp_models_folder = tmp_models_folder
        self.delete_tmp_models_folder_after_finish = delete_tmp_models_folder_after_finish
        self.delete_models_folder_after_finish = delete_models_folder_after_finish
        if n_jobs < 0:
            n_jobs = None
        self.n_jobs = n_jobs
        self.estimator = None
        super().__init__()

    def fit(self, **kwargs):
        """
        This the the whole step that the model could be trained using this function.
        args could be specified with sub-class.
        :param kwargs: key-words could be used to fit the model.
        :return: self
        """
        pass

    def predict(self, **kwargs):
        pass

    def score(self, **kwargs):
        pass

    @staticmethod
    def get_estimator():
        pass


class ClassificationEstimator(Estimators):
    def fit(self, xtrain,
            ytrain,
            xtest=None,
            ytest=None,
            n_jobs=None,
            batch_size=None):
        if n_jobs is None or n_jobs == 1:
            # with just one thread to do training
            # we have to instant the object.
            self.estimator = self.get_estimator()()
            # ensure data should be array type
            self.estimator.fit(xtrain, ytrain, xtest, ytest, batch_size)

        else:
            # we get n_jobs, we should use parallel training step.
            pass

    def predict(self, xtest, ytest, batch_size=None):
        pass

    def score(self, xtest, ytest, batch_size=None):
        pass

    @staticmethod
    def get_estimator():
        """
        this is to get the classification auto-ml object, so that
        we could use this object to do training.
        :return: class pointer
        """
        return ClassificationAutoML



# -*- coding:utf-8 -*-
"""
Parent for whole `networks` for this project that we could use for sub-class

@author: Guangqiang.lu
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import style

from sklearn.model_selection import train_test_split
from sklearn.exceptions import NotFittedError
import tensorflow as tf
from tensorflow.keras.callbacks import EarlyStopping

from auto_ml.utils.backend_obj import Backend
from auto_ml.utils.logger import create_logger

logger = create_logger(__file__)


class BaseNet(object):
    def __init__(self):
        self.backend = Backend()
        self.model = None

    def fit(self, data, label, epochs=100, batch_size=128, callback=None, patience=10):
        """
        Model fitting function, here this function will split data to be train and validation data sets.
        Model training will use train data, and will also evaluate after training model accuracy.
        :param model:
        :param data:
        :param label:
        :param n_classes:
        :param epochs:
        :param batch_size:
        :param callback:
        :param patience:
            how many epochs to wait for the model to train.
        :param silence:
        :return:
        """
        label = self._check_label(label)

        xtrain, xvalidate, ytrain, yvalidate = train_test_split(data, label, test_size=.2, random_state=1234)

        if callback is None:
            callback = EarlyStopping(monitor='val_loss', patience=patience, min_delta=1e-4)

        logger.info("Start to `train` neural network model based on training data!")
        self.his = self.model.fit(xtrain, ytrain, epochs=epochs, batch_size=batch_size,
                            validation_data=(xvalidate, yvalidate), callbacks=[callback], verbose=1)

        val_acc = self.model.evaluate(xvalidate, yvalidate, batch_size=1024)[1]
        logger.info('After training, model evaluate on validate data accuracy = {:.6f}%'.format(val_acc * 100))

        # As we need to save trained model into disk, we need the validation score combined with class name...so store
        # After the model is fitted, then let's plot the result, so we could get a better understanding.
        # self.plot_acc()

        return self.his

    def predict(self, data, batch_size=1024):
        """
        Get max index as prediction
        :param data:
        :param batch_size:
        :param step:
        :return:
        """
        prob = self.predict_proba(data, batch_size)

        if len(prob.shape) > 1 and len(prob) > 0:
            # just to ensure that we have get proper probability
            pred = np.argmax(prob, axis=1)
        else:
            raise ValueError("When to use neural network's `predict` func get error!")

        return pred

    def predict_proba(self, data, batch_size=1024):
        if not hasattr(self.model, 'predict'):
            # by default, keras will use `predict` to return the probability
            logger.error("Neural network's model hasn't been fitted, so please fit model first!")
            raise NotFittedError("Neural network's model hasn't been fitted, so please fit model first!")

        prob = self.model.predict(data, batch_size=batch_size)

        return prob

    def score(self, data, label, batch_size=1024):
        label = self._check_label(label)

        eval_acc = self.model.evaluate(data, label, batch_size=batch_size)[1]

        # To keep score with 6 digits to keep step with the whole framework.
        eval_acc = round(eval_acc, 6)
        logger.info('Model evaluate on `test data` accuracy = {:.6f}%'.format(eval_acc))

        return eval_acc

    def save(self, model_name, path=None):
        """
        Save model into disk.
        :param model_name:
        :param path:
        :return:
        """
        if path is None:
            # If path is not given, save model to default model path
            path = self.backend.output_folder

        if not model_name.endswith(".h5"):
            model_name += '.h5'

        self.model.save(os.path.join(path, model_name))
        logger.info('Model have been saved to disk: {}'.format(os.path.join(path, model_name)))

    def plot_acc(self, plot_acc=True, plot_loss=True, figsize=(8, 6)):
        """
        Currently is for `classification`.
        :param plot_acc:
        :param plot_loss:
        :param figsize:
        :return:
        """
        style.use('ggplot')

        if plot_acc:
            fig_1, ax_1 = plt.subplots(1, 1, figsize=figsize)
            ax_1.plot(self.his.history['accuracy'], label='Train Accuracy')
            ax_1.plot(self.his.history['accuracy'], label='Validation Accuracy')
            ax_1.set_title('Train & Validation Accuracy curve')
            ax_1.set_xlabel('Epochs')
            ax_1.set_ylabel('Accuracy score')
            plt.legend()

            plt.show()

        if plot_loss:
            fig_2, ax_2 = plt.subplots(1, 1, figsize=figsize)
            ax_2.plot(self.his.history['loss'], label='Train Loss')
            ax_2.plot(self.his.history['val_loss'], label='Validation Loss')
            ax_2.set_title('Train & Validation Loss curve')
            ax_2.set_xlabel('Epochs')
            ax_2.set_ylabel('Loss score')
            plt.legend()

            plt.show()

    # This is used for converting 1D label to nD one-hot label
    @staticmethod
    def _check_label(label):
        """
        Only use for `classification` problem.
        :param label:
        :return:
        """
        if len(label.shape) == 1:
            return tf.keras.utils.to_categorical(label, num_classes=len(np.unique(label)))

        return label



# -*- coding:utf-8 -*-
"""
Here is to select some features if we have so many features, one thing that we could
reduce our training time, also we could remove some relevant features.

@author: Guangqiang.lu
"""
from sklearn.feature_selection import VarianceThreshold, SelectFromModel
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import ExtraTreesClassifier
from auto_ml.preprocessing.processing_base import Process
from auto_ml.utils.data_rela import check_label
from auto_ml.utils.CONSTANT import *


class FeatureSelect(Process):
    def __init__(self, simple_select=True, tree_select=False):
        super(FeatureSelect, self).__init__()
        self.simple_select = simple_select
        self.tree_select = tree_select

    def fit(self, data, y=None):
        """
        Also support with algorithm based feature selection
        :param data: data to process
        :param label: label data if need with algorithm trained based
        :return:
        """
        if self.simple_select:
            self.estimator = VarianceThreshold()
            self.estimator.fit(data)
        else:
            # if we want to use algorithms based feature extraction, currently will use LinearRegression
            if y is None:
                raise ValueError("When want to use Algorithm, label data should be provided!")

            # before next step, we should ensure task should be classification
            label_type = check_label(y)
            if label_type not in CLASSIFICTION_TASK:
                raise ValueError("When we want to use model selection logic, task type should just be classification.")

            if self.tree_select:
                model = ExtraTreesClassifier(n_estimators=50).fit(data, y)
            else:
                model = LinearRegression().fit(data, y)

            self.estimator = SelectFromModel(model, prefit=True)


if __name__ == '__main__':
    from sklearn.datasets import load_iris

    x,y = load_iris(return_X_y=True)

    f = FeatureSelect(simple_select=False, tree_select=True)
    f.fit(x, y)
    print(f.transform(x))



# -*- coding:utf-8 -*-
"""
This is used to make test data for whole project

@author: Guangqiang.lu
"""
import os
import pandas as pd
from auto_ml.utils.paths import get_root_path
from auto_ml.utils.backend_obj import Backend


backend = Backend()
root_path = get_root_path()
cur_path = os.path.join(root_path, 'test')


def get_training_data(return_df=False, file_name='train.csv'):
    df = pd.read_csv(os.path.join(cur_path, file_name))

    if return_df:
        return df

    x = df.drop(['Survived'], axis=1).values
    y = df['Survived'].values
    x = x.copy(order='C')

    return x, y


def save_processing_data(data, dataset_name='process_tmp'):
    backend.save_dataset(data, dataset_name, model_file_path=False)
    print('Dataset has been saved into tmp folder.')


def load_processing_data(dataset_name='process_tmp'):
    """Load the processed data from disk"""
    dataset = backend.load_dataset(dataset_name)

    return dataset


# -*- coding:utf-8 -*-
"""
This is whole project logging functionality that could be used.

author: Guangqiang.lu
"""

import logging
import tempfile
import shutil
import os
from datetime import datetime

# Add with contextlib module to make the log to start and end message auto
from contextlib import contextmanager

from auto_ml.utils.CONSTANT import TMP_FOLDER

class Logger(object):
    def __init__(self, logger_name=None):
        if logger_name is None:
            logger_name = "logging"   # we could change this.
        self.logger_name = logger_name
        # self._logger_path = os.path.join(tempfile.gettempdir(), 'logging')
        if not os.path.exists(TMP_FOLDER):
            try:
                os.makedirs(TMP_FOLDER)
            except Exception as e: 
                raise IOError("When try to create TMP folder: {} get error: {}".format(TMP_FOLDER, e))
        self._logger_path = os.path.join(TMP_FOLDER, 'logging')
        if not os.path.exists(self._logger_path):
            try:
                os.mkdir(self._logger_path)
            except IOError as e:
                raise IOError("When try to create logging folder with error:%s" % e)

        self.logger = logging.getLogger(self.logger_name)
        self.logger.setLevel(logging.INFO)
        # this is to write logging info to disk
        # I just want to ensure there is only one logging file in current folder,
        # after the whole process finish, I just upload the file to HDFS for later use case
        self.logger_file_name = 'logging_%s.log' % datetime.now().strftime('%Y%m%d')
        self.logger_file_path = os.path.join(self._logger_path, self.logger_file_name)

        file_handler = logging.FileHandler(self.logger_file_path)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                                      datefmt='%Y-%m-%d %H:%M:%S')
        file_handler.setFormatter(formatter)
        # this is to write the logging info to console
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        self.logger.addHandler(file_handler)
        self._file_handler = file_handler

    def info(self, msg):
        self.logger.info(msg)

    def warning(self, msg):
        self.logger.warning(msg)

    def error(self, msg):
        self.logger.error(msg)

    def debug(self, msg):
        self.logger.debug(msg)

    def exception(self, msg):
        self.logger.exception(msg)

    def critical(self, msg):
        self.logger.critical(msg)

    def save_log_file(self, save_file_path):
        """
        To save the logger file into that folder,
        default will remove logger folder when call this func.
        :param save_file_path: Which folder to save
        :return:
        """
        try:
            # first we need to close the file writer
            self._file_handler.close()
            self.logger.removeHandler(self._file_handler)
        except Exception as e:
            raise IOError("When try to close file writer with error: {}".format(e))

        # then let's move the file into that folder
        try:
            shutil.move(self.logger_file_path, os.path.join(save_file_path, self.logger_file_name))
            # then just remove the tmp folder, as this function should be called only after whole step finish
            shutil.rmtree(self._logger_path)
        except IOError as e:
            raise IOError("When try to move logger file into {} with error: {}".format(save_file_path, e))


# def create_logger():
#     logger = Logger()

#     return logger

@contextmanager
def create_logger_context(logger_name):
    logger = Logger(logger_name)

    logger.info("Start logging for `{}`".format(logger_name))
    yield logger
    # logger.info("End logging for {}".format(logger_name))


def create_logger(logger_name=None):
    """Create a logger obj to log info for starting log"""
    if logger_name is None:
        logger_name = "logger"

    if os.path.isfile(logger_name):
        # in case that we provide with __file__ attr, then just with file name. 
        # so that we could get better insight.
        logger_name = os.path.basename(logger_name)

    with create_logger_context(logger_name) as logger:
        return logger

# logger = create_logger()

if __name__ == '__main__':
    logger = create_logger()
    logger.info("test")
    logger.warning("This is warning")

    logger2 = create_logger("test")
    logger2.info("new logger")


# -*- coding:utf-8 -*-
"""
This is a class that is used for creating auto-ml object that
we could use for training step.

author: Guangqiang.lu
"""

import numpy as np
import pandas as pd
import time
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import train_test_split

from auto_ml.utils.backend_obj import Backend
from auto_ml.base.classifier_algorithms import *
from auto_ml.metrics.scorer import *
from auto_ml.utils.CONSTANT import *
from auto_ml.pipelines.pipeline_training import ClassificationPipeline
from auto_ml.utils.logger import create_logger


logger = create_logger(__file__)


class AutoML(BaseEstimator):
    def __init__(self, models_path=None,
                 time_left_for_this_task=3600,
                 n_ensemble=10,
                 n_best_model=5,
                 include_estimators=None,
                 exclude_estimators=None,
                 include_preprocessors=None,
                 exclude_preprocessors=None,
                 keep_models=True,
                 model_dir=None,
                 precision=32,
                 delete_models=True
                 ):
        """
        this is to init automl class, whole thing should be ininstanted
        in this class, like what algorithms to use, how many models to
        be selected, etc.
        :param backend: backend object used to save and load models
        :param time_left_for_this_task: how long for this models to be trained.
        :param n_ensemble: how many models to be selected to be ensemble
        :param n_best_model: how many models to be keeped during training.
        :param include_estimators: what algorithms to be included
        :param exclude_estimators: what algorithms to be excluded
        :param include_preprocessors: what preprocessing step to be included
        :param exclude_preprocessors: what preprocessing step to be excluded
        :param keep_models: whether or not to keep trained models
        :param model_dir: keep model folder, if None use backend to create one folder
        :param precision: precision of data, to save memory
        """
        super(AutoML, self).__init__()
        # Change backend object creation as this class is entry class also Backend class support singleton, so output_folder will work.
        models_path = OUTPUT_FOLDER if models_path is None else models_path
        # Add with `models_path` parameters, as the only output for the framework is only models files, so to store models into
        # folder that we would like will be nice!
        self.models_path = models_path
        
        self.backend = Backend(output_folder=models_path) 
        self.time_left_for_this_taks = 3600 if time_left_for_this_task is None else time_left_for_this_task
        self.n_ensemble = n_ensemble
        self.n_best_model = n_best_model
        self.include_estimators = include_estimators
        self.exclude_estimators = exclude_estimators
        self.include_preprocessors = include_preprocessors
        self.exclude_preprocessors = exclude_preprocessors
        self.keep_models = keep_models
        self.model_dir = model_dir
        self.precision = precision

        self.estimator = None

        # as we will use `ensemble` to combine models so the last will just be one model
        self.best_model = None

        # Add with what type of the problem
        self.type_of_problem = None

        self.delete_models = delete_models

    def fit(self, *args, **kwargs):
        """
        Type of the problem attribute should be added, so that for `score`,
        we could get metrics based on different problem.
        :param xtrain:
        :param ytrain:
        :param n_jobs:
        :return:
        """
        raise NotImplementedError

    def predict(self, x, **kwargs):
        """
        Most of the prediction logic should happen here, as for the score should based on
        prediction result.

        :param x:
        :param kwargs:
        :return:
        """
        logger.info("Start to get prediction based on best trained model.")
        prediction = self.estimator.predict(x)
        logger.info("Prediction step finishes.")

        return prediction

    def predict_proba(self, x, **kwargs):
        """
        Should support with probability if supported.
        :param x:
        :param kwargs:
        :return:
        """
        # should check the estimator should have function: `predict_proba`!
        if not hasattr(self.estimator, 'predict_proba'):
            raise NotImplementedError("Best fitted model:{} doesn't support `predict_proba`".format(self.best_model))

        logger.info("Start to get probability based on best trained model.")
        prob = self.estimator.predict_proba(x)
        logger.info("Prediction step finishes.")

        return prob

    def score(self, x, y, **kwargs):
        self._check_fitted()

        logger.info("Start to get prediction based on best trained model!")

        # Use child func.
        score = self.estimator.score(x, y)

        logger.info("Get score: {} based on best model!".format(score))

        return score

    def _load_trained_models_ordered_by_score(self, higher_best=True):
        """
        To load whole trained model from disk and sorted them based on `higher_best`.
        :param higher_best:
        :return:
            models_list: a list of trained models sorted with score. [("lr-0.9.pkl", lr-0.9.pkl), ...]
        """
        models_list = self.backend.load_models_combined_with_model_name()

        # let's sort these models based on score, this is a tuple:(model_name, model_instance)
        if higher_best:
            reverse = True
        else:
            reverse = False

        # sort models by `training` score with diff type of problem
        models_list = sorted(models_list,
                             key=lambda model: float(model[0].split("-")[1].replace(".pkl", '').replace(".h5", '')),
                             reverse=reverse)

        return models_list

    def get_sorted_models_scores(self, x, y, **kwargs):
        """
        Add this func to get whole score based for each trained models, so that
        we could get the result that we have taken that times and for each models,
        how about the testing result.

        But this should be implemented by pipeline!
        :param x:
        :param y:
        :param kwargs:
        :return:
            a list of tuple: [(model_name, model_score), ...]
        """
        raise NotImplementedError

    def _check_fitted(self):
        if not self.models_list:
            logger.error("When try to get prediction with `automl`, Please fit the model first")
            raise NotFittedError("When try to get prediction with `automl`, Please fit the model first")

        return True


class ClassificationAutoML(AutoML):
    def __init__(self, models_path=None, 
            include_estimators=None, 
            exclude_estimators=None, 
            include_preprocessors=None, 
            exclude_preprocessors=None,
            **kwargs):
        """Added with algorithm selection and processing selection, even with others in case we need.

        Args:
            models_path (Str, optional): Where to store our models. Defaults to None.
        """
        super(ClassificationAutoML, self).__init__(models_path=models_path,
                                            include_estimators=include_estimators,
                                            exclude_estimators=exclude_estimators,
                                            include_preprocessors=include_preprocessors,
                                            exclude_preprocessors=exclude_preprocessors, **kwargs)
        # after pipeline has finished, then we should use `ensemble` to combine these models
        # action should happen here.
        self.estimator = ClassificationPipeline(backend=self.backend,
                                            include_estimators=include_estimators,
                                            exclude_estimators=exclude_estimators,
                                            include_preprocessors=include_preprocessors,
                                            exclude_preprocessors=exclude_preprocessors, 
                                            **kwargs)

    def fit(self, x=None, y=None, file_load=None, \
             xval=None, yval=None, val_split=None, n_jobs=None, use_neural_network=True):
        """
        Real training logic happen here, also store trained models.

        Support both with file to train also with data and label data.

        :param xtrain: train data
        :param ytrain: label data
        :param n_jobs: how many cores to use
        :return:
        """
        start_time = time.time()

        # Add logic here is that we should clean models' folder, so that every time we could get a clean folder for next time running!
        if self.backend and self.delete_models:
            self.backend.clean_folder()

        x, y = self._get_data_and_label(file_load, x, y)

        if val_split is not None:
            # if do need to do validation based on current train data, then just split current data into validation as well
            x, xval, y, yval = train_test_split(x, y, test_size=val_split)
        else:
            # Here I think if and only if the data length is over a threashold, then to do validation,even user haven't provided val data.
            if len(x) > VALIDATION_THRESHOLD:
                val_split = .2
                # if do need to do validation based on current train data, then just split current data into validation as well
                x, xval, y, yval = train_test_split(x, y, test_size=val_split)

        # after the checking process, then we need to create the Pipeline for whole process.
        # Here should use a Pipeline object to do real training, also with `ensemble`
        self.estimator.fit(x, y, n_jobs=n_jobs, use_neural_network=use_neural_network)

        # load trained models for prediction and scoring for testing data.
        # after we have fitted the trained models, then next step is to load whole of them from disk
        # and sort them based on score, and get scores for `test data` and `test label`
        # WE could get them from parent function, so that we could also use this for `regression`
        self.models_list = self._load_trained_models_ordered_by_score(higher_best=True)

        self._validation_models(xval, yval)

        logger.info("Whole training pipeline takes: {} seconds!".format(round(time.time() - start_time, 2)))

    @staticmethod
    def _get_data_and_label(file_load, x, y):
        """
        Ensure could get data and label.
        """
        if file_load is None and x is None and y is None:
            raise ValueError("When do real training, please provide at least a " +
                 "`file_load` or train data with `xtrain, ytrain`!")

        if file_load is not None:
            # with container, then just query the attribute then we could keep other as same.
            x, y = file_load.data, file_load.label
        else:
            if x is None or y is None:
                raise ValueError("When to do training, please provide both `x` and `y`!")
        
        return x, y
    
    def _validation_models(self, xval, yval):
        if xval is not None and yval is not None:
            score_dict = self.estimator.get_sorted_models_scores(xval, yval)
            print(score_dict)
            
            score_log_str = self.__format_trained_model_scores(score_dict)
            print(score_log_str)
        else:
            logger.warning("No need to validation!")
            pass

    @staticmethod
    def __format_trained_model_scores(score_dict, n_space=35):
        out_str_format = '{{0:{0}}}|{{1:{0}}}|{{2:{0}}}'.format(n_space)

        score_log_str = out_str_format.format("Model name", "Train score", "Validation score")
        logger.info(score_log_str)

        for model_name, test_score in score_dict.items():
            model_name_split = model_name.split('-')
            # model_name = model_name_split[0]
            train_score = model_name_split[1]
            train_score = train_score[:train_score.rindex(".")]
            # must convert to str, otherwise with not wanted result.
            test_score = str(test_score)
            
            log_str = out_str_format.format(model_name, train_score, test_score)
            score_log_str += '\n' + log_str

            logger.info(log_str)
        
        return score_log_str

    def get_sorted_models_scores(self, file_load=None, xtest=None, ytest=None, **kwargs):
        """
        To get some best trained model's score for `test` data with ordered.

        So that we could get the list of the best scores for later front end show case.
        :param x:
        :param y:
        :param kwargs:
        :return:
        """
        self._check_param(file_load, xtest, ytest)

        if file_load is not None:
            xtest, ytest = file_load.data, file_load.label

        score_dict = self.estimator.get_sorted_models_scores(xtest, ytest, reverse=True)

        return score_dict

    def predict(self, x=None, file_load=None, **kwargs):
        """To support with file_load obj with super func."""
        self._check_param(file_load, x)

        if file_load is not None:
            x, _ = self.__get_file_load_data_label(file_load)

        pred = super().predict(x)

        return pred

    def predict_proba(self, x=None, file_load=None, **kwargs):
        self._check_param(file_load, x)

        if file_load is not None:
            x, _ = self.__get_file_load_data_label(file_load)

        prob = super().predict_proba(x)

        return prob

    def score(self, x=None, y=None, file_load=None, **kwargs):
        self._check_param(file_load, x, y)

        if file_load is not None:
            x, y = self.__get_file_load_data_label(file_load, use_for_pred=False)
        
        score = super().score(x, y)

        return score

    @staticmethod
    def _check_param(*args):
        all_None = all([arg is None for arg in args])

        if all_None:
            raise ValueError("Please provide at least one parameter!")

    @staticmethod
    def __get_file_load_data_label(file_load, use_for_pred=True):
        """
        Get data and label for prediction data.
        """
        data, label = file_load.data, file_load.label

        if use_for_pred:
            if label is not None:
                raise ValueError("When to `predict` with `file_load` obj, if you set parameter: `use_for_pred=True`, then please provide with `label` column." + 
                        "As we couldn't get label data from file_load obj.")
        
        return data, label

    @classmethod
    def reconstruct(cls, models_path=None, *args, **kwargs):
        """
        Used for Restful API to create
        """
        return cls(models_path, *args, **kwargs)

        

class FileLoad:
    """Load data from file, support with local file also with GCS.

    Make this class as a container for later use case.
    """
    def __init__(self, file_name, file_path=None, file_sep=',', label_name='label', use_for_pred=False,
            service_account_file_name=None, service_account_file_path=None, except_columns=None):
        """
        Support with prediction data load. If true, have to set parameter: `use_for_pred=True`
        """
        if not file_name.endswith('csv'):
            raise ValueError("Currently only support CSV file! Please provide with a CSV file.")
        self.file_name = file_name
        self.file_sep = file_sep
        self.file_path = file_path if file_path is not None else '.'
        self.label_name = label_name
        self.use_for_pred = use_for_pred
        self.service_account_file_name = service_account_file_name
        self.service_account_file_path = service_account_file_path
        if except_columns is not None and isinstance(except_columns, str):
            except_columns = [except_columns]
        self.except_columns = except_columns 
        self.data, self.label = self._load_data_file()

    def _get_file_location(self):
        if self.file_path is not None:
            if self.file_path.startswith("gs://"):
                return 'gcs'
            else:
                return 'local'

    def _get_gcs_file(self):
        # first service account
        self._check_service_account()

        bucket_name = self.file_path.split('/')[2]
        if bucket_name.find(".") > 0:
            # file_path also contain file name with `.`
            logger.error("When to get bucket name from file path: {}, couldn't get the bucket name.".format(self.file_path))
            raise ValueError("When to get bucket name from file path: {}, couldn't get the bucket name.".format(self.file_path))

        from google.cloud import storage
        
        # init client with service account file
        client = storage.Client.from_service_account_json(os.path.join(self.service_account_file_path, self.service_account_file_name))

        bucket = client.bucket(bucket_name)
        blob = bucket.blob(self.file_name)

        try:
            # download as a file into temperate folder
            blob.download_to_filename(os.path.join(TMP_FOLDER, self.file_name))
            logger.info("File: {} has been downloaded from GCS to local.".format(self.file_name))

        except Exception as e:
            logger.error("Error downloading file with service account: {}".format(self.service_account_file_name))
            raise e
    
    def _check_service_account(self):
        # Check service account related
        if not self.service_account_file_name:
            raise ValueError("When try to use GCS, please must provide with service account file to interact with GCS!")

        if not self.service_account_file_name.endswith('json'):
            raise ValueError("When try to use GCS, Service acount file must be a JSON file, " +\
                 "but provided with {}".format(self.service_account_file_name))
                
        if not os.path.exists(os.path.join(self.service_account_file_path, self.service_account_file_name)):
            raise FileNotFoundError("When try to find service account file couldn't find file: {} in path: {}, " + 
                    "please check it.".format(self.service_account_file_name, self.service_account_file_path))
    
    def _load_data_file(self):
        file_location = self._get_file_location()
        
        file_path = self.file_path

        if file_location == 'gcs':
            logger.info("Start to download file:{} from GCS.".format(self.file_name))
            self._get_gcs_file()
            file_path = TMP_FOLDER 
            logger.info("Download file from GCS has finished.")
        
        try:
            df = pd.read_csv(os.path.join(file_path, self.file_name), sep=self.file_sep)

            df_cols = df.columns

            # In case that we don't need some columns and these cols should be in DF columns
            if self.except_columns is not None:
                except_common_cols = set(self.except_columns).intersection(list(df.columns))
                
                logger.info("Need to delete columns:{}  from original DataFrame.".format('\t'.join(list(except_common_cols))))

                if except_common_cols:
                    df.drop(except_common_cols, axis=1, inplace=True)

            if self.use_for_pred:
                # if just for prediction, then we don't get label data
                data = df.values
                label = None

                return data, label

            # let's inference data and label from DataFrame
           
            if self.label_name not in df_cols:
                # if `label` col not in the original file, then make let first column as `label` column
                label = df.iloc[:, 0]
                data = df.iloc[:, 1:]
            else:
                label = df[self.label_name]
                data = df.drop([self.label_name], axis=1)
            
            # convert into array type will be easier 
            data, label = data.values, label.values.reshape(-1, 1)
            
            return data, label
        except IOError as e:
            logger.error("Error load data file: {} from path: {}".format(self.file_name, file_path))
            raise e
        

if __name__ == '__main__':
    # Test with `FileLoad` class
    file_name = 'train.csv'
    file_path = r"C:\Users\guangqiiang.lu\Documents\lugq\code_for_future\auto_ml_pro\auto_ml\test"
    # file_path = "gs://cloud_sch_test"
    # service_account_file_path = r"C:\Users\guangqiiang.lu\Downloads"
    # service_account_name = "buoyant-sum-302208-4542dcd74629.json"

    # file_load = FileLoad(file_name, file_path, file_sep=',',  label_name='Survived', 
    #     service_account_file_name=service_account_name, service_account_file_path=service_account_file_path)
        
    file_load = FileLoad(file_name, file_path, file_sep=',',  label_name='Survived')
    models_path = r"C:\Users\guangqiiang.lu\Downloads\test_automl"

    auto_cl = ClassificationAutoML(models_path=models_path)

    # Start to train processing for `FileLoad`
    # auto_cl.fit(file_load=file_load)
    # print(auto_cl.models_list)
    # # print(auto_cl.score(file_load_pred))

    # file_load_pred = FileLoad("test.csv", file_path, file_sep=',', use_for_pred=True)
    # print('*' * 20)
    # print(auto_cl.predict(file_load=file_load_pred)[:10])
    # print('*'*20)
    # print(auto_cl.predict_proba(file_load=file_load_pred)[:10])

    # try to use sklearn iris dataset
    from sklearn.datasets import load_iris
    from sklearn.model_selection import train_test_split
    
    x, y = load_iris(return_X_y=True)
    xtrain, xtest, ytrain, ytest = train_test_split(x, y, test_size=.2)
    
    auto_cl.fit(xtrain, ytrain)

    print(auto_cl.models_list)
    print(auto_cl.score(xtest, ytest))
    print('*' * 20)
    print(auto_cl.predict(xtest)[:10])
    print('*'*20)
    print(auto_cl.predict_proba(xtest)[:10])

    # # get model score
    # print(auto_cl.get_sorted_models_scores(xtest, ytest))

    # This is used for the submition for Kaggle
    # pred_data = pd.read_csv(os.path.join(file_path, test_file_name))

    # pass_id_df = pred_data[['PassengerId']]

    # prediction = auto_cl.predict(file_load_pred)
    # pred_df = pd.DataFrame(prediction, columns=['Survived'])

    # pred_df = pd.concat([pass_id_df, pred_df], axis=1)

    # pred_df.to_csv(os.path.join(file_path,"Submition_with_automl.csv"), index=False)




"""
This is used for some functions utils that we could use for the whole project.

"""
import warnings
import functools


def deprecated(func):
    """Used to deprecated some functions that we don't need, with a warning raised."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn("Called a function that we don't need: {}".format(func.__name__),
            category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)
        return func(*args, **kwargs)
    return new_func




# -*- coding:utf-8 -*-
from tensorflow.keras.layers import Dense, Dropout, BatchNormalization, Input, Activation, LSTM, Bidirectional, GRU, RNN
from tensorflow.keras.models import Model

from auto_ml.nueral_networks.basic_net import BaseNet


class RnnNet(BaseNet):
    """This class implements recurrent neural networks including basic RNN, LSTM, GRU, Bidirectional LSTM.
        Arguments:
            n_classes(Integer): How many classes you want to classifier or use it for regression.
                Integer 2 is for binary problem, Integer >2 is for how many classes you have,
                Integert -1 is for regression problem.

            input_dim_1(Integer): First Dimension of your data sets. Noted, this is not batch-size dimension.

            input_dim_2(Integer): Second Dimension of your data sets.

            input_dim_3(Integer, optional): Third Dimension of your data sets. Noted, if data sets is just 2D,
                this parameter should not be given.

            n_layers(Integer, optional): How many recurrent layers wanted to be used.

            use_dropout(Boolean, optional): Whether or not to use Dropout for model training, avoiding model overfitting.

            drop_ratio(Double, optional): How many ratios to drop units in each layers.

            use_lstm(Boolean, optional): Whether or not to use LSTM units as basic recurrent.
                Default is True, use LSTM.

            use_bidirec(Boolean, optional): Whether or not to use bidirectional recurrent to train.
                Default is Flase, if set True, then if 'use_lstm' is also True, then use LSTM units as basic,
                if 'use_gru' is True, then use GRU units as basic, else just use RNN units as basic.

            use_gru(Boolean, optional): Whether or not to use GRU units as basic recurrent.

            rnn_units(Integer, optional): How many units to be used as recurrent units.

            use_dense(Boolean, optional): Whether or not to use Dense layers.
                Default is True

            dense_units(Integer, optional): How many units for building dense layers.

            use_batch(Boolean, optional): Whether or not to use BatchNormalization to normalize model weights.

            loss(String, optional): Which loss function to be used for model training.
                Default is None, use default loss function. For binary, use 'binary_crossentropy';
                For multi-class, use 'categorical_crossentropy'; For regression, use 'mse'.

            metrics(String, optional): Which metric to be used for model evaluation.

            optimizer(String, optional): Which optimizer to be used for model training optimizer.

            silence(Boolean, optional): Whether to print some useful information to console.
                Default is False, means print something to console.
    """

    def __init__(self, n_classes, input_dim_1, input_dim_2, input_dim_3=None, n_layers=3, use_dropout=True, drop_ratio=.5,
                 use_lstm=True, use_bidirec=False, use_gru=False, rnn_units=64, use_dense=True, dense_units=64, use_batch=True,
                 loss=None, metrics='accuracy', optimizer='rmsprop', silence=False):
        super().__init__()
        self.n_classes = n_classes
        self.input_dim_1 = input_dim_1
        self.input_dim_2 = input_dim_2
        self.input_dim_3 = input_dim_3
        self.n_layers = n_layers
        self.use_dropout = use_dropout
        self.drop_ratio = drop_ratio
        self.use_lstm = use_lstm
        self.use_bidierc = use_bidirec
        self.use_gru = use_gru
        self.rnn_units = rnn_units
        self.use_dense = use_dense
        self.use_batch = use_batch
        self.dense_units = dense_units
        self.loss = loss
        self.metrics = metrics
        self.optimizer = optimizer
        self.silence = silence
        self.model = self._init_model()

    # this basic RNN block construction function.
    def _rnn_block(self, layers, name_index=None):
        if self.use_bidierc:
            if self.use_lstm:
                res = Bidirectional(LSTM(self.rnn_units, return_sequences=True, recurrent_dropout=self.drop_ratio))(layers)
            elif self.use_gru:
                res = Bidirectional(GRU(self.rnn_units, return_sequences=True, recurrent_dropout=self.drop_ratio))(layers)
            else:
                res = Bidirectional(RNN(self.rnn_units, return_sequences=True))(layers)

        elif self.use_gru:
            res = GRU(self.rnn_units, return_sequences=True,
                      recurrent_dropout=self.drop_ratio)(layers)

        elif self.use_lstm:
            res = LSTM(self.rnn_units, return_sequences=True,
                       recurrent_dropout=self.drop_ratio)(layers)
        else:
            res = RNN(self.rnn_units, return_sequences=True)(layers)

        if self.use_dropout:
            res = Dropout(self.drop_ratio)(res)

        return res

    def _init_model(self):
        inputs = Input(shape=(self.input_dim_1, self.input_dim_2))

        # No matter for LSTM, GRU, bidirection LSTM, final layer can not use 'return_sequences' output.
        for i in range(self.n_layers - 1):
            if i == 0:
                res = self._rnn_block(inputs, name_index=i)
            else:
                res = self._rnn_block(res, name_index=i)

        # final LSTM layer
        if self.use_bidierc:
            res = Bidirectional(LSTM(self.rnn_units))(res)
        elif self.use_gru:
            res = GRU(self.rnn_units)(res)
        elif self.use_lstm:
            res = LSTM(self.rnn_units)(res)
        else:
            res = RNN(self.rnn_units)(res)

        # whether or not to use Dense layer
        if self.use_dense:
            res = Dense(self.dense_units)(res)
            if self.use_batch:
                res = BatchNormalization()(res)
            res = Activation('relu')(res)
            if self.use_dropout:
                res = Dropout(self.drop_ratio)(res)

                # this is method private function to check whether or not loss is not given, then use default loss

        def _check_loss(model, loss, metrics, optimizer):
            if loss is not None:
                model.compile(loss=loss, metrics=[metrics], optimizer=optimizer)
            return model

        if self.n_classes == 2:  # this is binary class problem.
            out = Dense(self.n_classes, activation='sigmoid')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='binary_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes >= 2:  # this is multiclass problem.
            out = Dense(self.n_classes, activation='softmax')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='categorical_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes == -1:  # this is regression problem
            out = Dense(1)(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='mse', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        else:
            raise AttributeError("Parameter 'n_classes' should be -1, 2 or up 2!")

        if not self.silence:
            print('Model structure summary:')
            model.summary()

        return model




# -*- coding:utf-8 -*-
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Conv1D, BatchNormalization, Dense, Dropout, Activation, \
    GlobalAveragePooling1D, Flatten, MaxPooling1D, MaxPooling2D, Conv2D, GlobalAveragePooling2D, add, concatenate

from auto_ml.nueral_networks.basic_net import BaseNet


class ResidualNet(BaseNet):
    """This class implements basic residualNet, also with DenseNet.
    
        Arguments:
            n_classes(Integer): How many classes you want to classifier or use it for regression.
                Integer 2 is for binary problem, Integer >2 is for how many classes you have,
                Integert -1 is for regression problem.

            input_dim_1(Integer): First Dimension of your data sets. Noted, this is not batch-size dimension.

            input_dim_2(Integer): Second Dimension of your data sets.

            input_dim_3(Integer, optional): Third Dimension of your data sets. Noted, if data sets is just 2D,
                this parameter should not be given.

            is_2D(Boolean, optional): Whether or not data sets is just 2D.
                Default is True.

            is_3D(Boolean, optional): Whether or not data sets is just 3D.
                Default is False.
                
            basic_residual(Boolean, optional): Whether to use basic ResidualNet.
                Default is True, if set this param to False, then will use DenseNet.

            n_res_layers(Integer, optional): How many residual layers wanted to construct.

            conv_units(Integet, optional): How many units for each Convolutional layer.
                For better accuracy, give more units will be good.

            kernel_size(Integet, optional): How big is kernel wanted to slide on data sets.
                This kernel will be (n*n)

            strides(Integer, optional): How many steps make the kernel slide on data sets.

            padding(String, optional): Whether or not drop or fill the missing values.
                Default is 'SAME': returned result is same as before, if 'valid': missing value will be droped.

            use_global(Boolean, optional): Whether or not to use global average pooling.
                Default is False, meaning Flatten Conv result.

            use_pooling(Boolean, optional): Whether or not to use pooling after Convolutional layer.
                Default if False, meaning without Pooling layer.

            use_concat(Boolean, optional): Whether to use 'concat' or 'add' to merge layers.
                Default is False use 'add' to merge layers.

            use_dropout(Boolean, optional): Whether or not to use Dropout after each layer.
                Default is True, recommended when model is overfitting.

            drop_ratio(Double, optional): How many ratios to drop units during training.
                Default is 0.5, if wanted for bigger penalty, set this bigger, may cause underfitting.

            use_batch(Boolean, optional): Whether or not to use BatchNormalization durng training.
                Default is True, this is great for improving model training accuracy

            activation(String, optional): Which activation function wanted to be used.
                Default is 'relu'

            use_dnn(Boolean, optional): Whether or not to use the Dense layer.
                Default is True

            n_dnn_layers(Integer, optional): How many dense layer wanted to be used.

            dnn_units(Integer, optional): How many units to be used for each layer.

            loss(String, optional): Which loss function wanted to be used during training as objective function.
                Default is None, it will be choosen according to different problems. For 'binary' problem, use
                'binary_crossentropy', for 'multi-class' problem, use 'categorical_crossentropy', for 'regression'
                problem, use 'mse'.

            metrics(String, optional): Which metric wanted to be used during training.

            optimizer(String, optional): Which optimizer to be used for optimizing.

            silence(Boolean, optional): Whether to print some useful information to console.
                Default is False, means print something to console.
    
    """
    def __init__(self, n_classes, input_dim_1, input_dim_2, input_dim_3=None,  is_2D=True, is_3D=False,
                 basic_residual=False, n_res_layers=4, kernel_size=2, stridess=1, padding='SAME',
                 use_global=False, use_dense=True, use_pooling=False, use_concat=False,
                 n_dnn_layers=1, conv_units=64, dnn_units=128, use_dropout=True,
                 drop_ratio=.5, use_batch=True, loss=None, optimizer='rmsprop', metrics='accuracy', silence=False):
        super(ResidualNet, self).__init__()
        self.input_dim_1 = input_dim_1
        self.input_dim_2 = input_dim_2
        self.input_dim_3 = input_dim_3
        self.is_2D = is_2D
        self.is_3D = is_3D
        self.n_classes = n_classes
        self.basic_residual = basic_residual
        self.n_res_layers = n_res_layers
        self.use_global = use_global
        self.use_dense = use_dense
        self.use_pooling = use_pooling
        self.use_concat = use_concat
        self.n_dnn_layers = n_dnn_layers
        self.conv_units = conv_units
        self.kernel_size = kernel_size
        self.strides = stridess
        self.padding = padding
        self.dnn_units = dnn_units
        self.use_dropout = use_dropout
        self.drop_ratio = drop_ratio
        self.use_batch = use_batch
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
        self.silence = silence
        self.model = self._init_model()

    # Residual block including for using basic ResidualNet or use DenseNet.
    # This including 2 convolutional layers in each block.
    def _res_block(self, layers, added_layers):
        # Is 2D or 3D data, also with whether or not to use pooling layer.
        if self.dimstr == '2D':
            res = Conv1D(self.conv_units, self.kernel_size, self.strides, padding=self.padding)(layers)
            if self.use_pooling:
                res = MaxPooling1D()(res)
        elif self.dimstr == '3D':
            res = Conv2D(self.conv_units, self.kernel_size, self.strides, padding=self.padding)(layers)
            if self.use_pooling:
                res = MaxPooling2D()(res)

        if self.use_batch:
            res = BatchNormalization()(res)
        res = Activation('relu')(res)
        if self.use_dropout:
            res = Dropout(self.drop_ratio)(res)

        # Is 2D or 3D data
        if self.dimstr == '2D':
            res = Conv1D(self.input_dim_2, self.kernel_size, self.strides, padding=self.padding)(res)
        elif self.dimstr == '3D':
            res = Conv2D(self.input_dim_2, self.kernel_size, self.strides, padding=self.padding)(res)

        if self.use_batch:
            res = BatchNormalization()(res)
        res = Activation('relu')(res)
        if self.use_dropout:
            res = Dropout(self.drop_ratio)(res)

        # This is basic ResidualNet.
        if self.basic_residual:
            if self.use_concat:
                return concatenate([res, layers])
            else:
                return add([res, layers])
        # This is DenseNet.
        else:
            if self.use_concat:
                return concatenate([res, added_layers])
            else:
                return add([res, added_layers])

    # This will build DenseNet or ResidualNet structure, model is already compiled.
    def _init_model(self):
        # Check the input parameters for getting data shape is '2D' or '3D'
        self.dimstr = check_dims(self.input_dim_3, self.is_2D, self.is_3D)
        inputs = Input(shape=(self.input_dim_1, self.input_dim_2))

        # construct residual block chain.
        for i in range(self.n_res_layers):
            if i == 0:
                res = self._res_block(inputs, inputs)
            else:
                res = self._res_block(res, inputs)

        # using flatten or global average pooling to process Convolution result
        if not self.use_global:
            res = Flatten()(res)
        else:
            if self.dimstr == '2D':
                res = GlobalAveragePooling1D()(res)
            elif self.dimstr == '3D':
                res = GlobalAveragePooling2D()(res)

        # whether or not use dense net, also with how many layers to use
        if self.use_dense:
            for _ in range(self.n_dnn_layers):
                res = Dense(self.dnn_units)(res)
                if self.use_batch: res = BatchNormalization()(res)
                res = Activation('relu')(res)
                if self.use_dropout: res = Dropout(self.drop_ratio)(res)

        # check parameter 'loss' is given or not, if not, use default loss for model training.
        def _check_loss(model, loss, metrics, optimizer):
            if loss is not None:
                model.compile(loss=loss, metrics=[metrics], optimizer=optimizer)
            return model

        if self.n_classes == 2:  # this is binary class problem.
            out = Dense(self.n_classes, activation='sigmoid')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='binary_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes >= 2:  # this is multiclass problem.
            out = Dense(self.n_classes, activation='softmax')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='categorical_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes == -1:  # this is regression problem
            out = Dense(1)(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='mse', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        else:
            raise AttributeError("Parameter 'n_classes' should be -1, 2 or up 2!")

        if not self.silence:
            print('Model structure summary:')
            model.summary()

        return model



"""Search a best neural network model with `kerastuner` based on data we have.

"""
import os
import traceback
import time
import string
import random
import shutil
import warnings
warnings.simplefilter("ignore")

from sklearn.model_selection import train_test_split

# This won't work as check `keras-tuner` source code that uses `print` to get info...
# # This only workable in linux
# os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
# This is to filter tensorflow warning log
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

from tensorflow import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, LSTM, Dropout, Conv2D, Conv1D, Activation
from kerastuner import HyperModel
from kerastuner.tuners import RandomSearch


from auto_ml.utils.paths import load_yaml_file
from auto_ml.utils.CONSTANT import OUTPUT_FOLDER, TMP_FOLDER
from auto_ml.utils.logger import create_logger
from auto_ml.utils.data_rela import get_num_classes_based_on_label


logger = create_logger(__file__)


hyper_yml_file_name = 'search_hypers.yml'
# Load hyperparameters
hyper_yml = load_yaml_file(hyper_yml_file_name)


# To build the search model based on the hyper yaml file
class SearchModel(HyperModel):
    def __init__(self, n_classes, algorithm_name='DNN', type_of_problem='classification', use_dropout=True):
        self.n_classes = n_classes
        self.algorithm_name = algorithm_name
        self.type_of_problem = type_of_problem
        self.use_dropout = use_dropout
        self.n_layers = hyper_yml[algorithm_name]['n_layers']
        self.n_units = hyper_yml[algorithm_name]['n_units']
        self.optimizers = hyper_yml['optimizers']
        self.hp = None
    
    def build(self, hp):
        raise NotImplementedError

    def _compile_model(self, model, optimizer_name, 
            loss_name='sparse_categorical_crossentropy', 
            metrics='accuracy', 
            type_of_problem='classification'):

        learning_optimizer = self._build_search_choice('learning_rate', self.optimizers[optimizer_name])
        
        if type_of_problem == 'regression':
            loss_name = 'mse'
            metrics = None

        # default to make with `Adam`
        optimizer = keras.optimizers.Adam(learning_optimizer)
        if optimizer_name == 'SGD':
            optimizer = keras.optimizers.SGD(learning_optimizer)
        
        compile_dict = {'optimizer': optimizer, "loss":loss_name, 'metrics': [metrics]}

        model.compile(**compile_dict)

        return model

    def _build_search_range(self, param_range, name=None):
        return self.hp.Int(name, param_range['min_value'], param_range['max_value'], param_range['step'])

    def _build_search_choice(self, name, param_range):
        return self.hp.Choice(name, param_range[name])



class DNNSearch(SearchModel):
    def __init__(self, n_classes, type_of_problem='classification', use_dropout=False):
        super().__init__(n_classes, algorithm_name='DNN', type_of_problem=type_of_problem, use_dropout=use_dropout)
        self.name = "DNN"

    def build(self, hp):
        self.hp = hp

        model = Sequential()

        for i in range(self._build_search_range(self.n_layers, 'n_layers')):
            model.add(Dense(units=self._build_search_range(self.n_units, 'n_units_' + str(i)), activation='relu'))
            if self.use_dropout:
                model.add(Dropout(0.5))
        
        model.add(Dense(self.n_classes))

        model = self._compile_model(model, 'Adam', type_of_problem=self.type_of_problem)
        return model


class EvaluateNeuralModel:
    def __init__(self, model_list, x, y, algorithm_name='DNN'):
        self.model_list= model_list if isinstance(model_list, list) else [model_list]
        self.x = x
        self.y = y
        self.algorithm_name = algorithm_name
    
    def evaluate_models(self):
        self._check_models()

        score_list = []
        for model in self.model_list:
            try:
                score = model.evaluate(self.x, self.y)[1]
                # score will be with 6 digits
                score = round(score, 6)

                score_list.append(score)
            except ValueError as e:
                raise ValueError("When try to evaluate model with self data get error: {}".format(e))
            
        return score_list
    
    def save_models(self, model_path, model_name_suffix='.h5'):
        score_list = self.evaluate_models()

        if len(self.model_list) != len(score_list):
            raise ValueError("When to save model into disk, got score list is not equal to model list. \
                Model list number: {}, score list number:{}".format(len(self.model_list), len(score_list)))

        if not model_path:
            model_path = OUTPUT_FOLDER

        for model, model_score in zip(self.model_list, score_list):
            model_name = self.algorithm_name + '-' + str(model_score) + model_name_suffix
            try:
                model.save(os.path.join(model_path, model_name))

                logger.info("Model: {} bas been save into folder: {}".format(model_name, model_path))
            except Exception as e:
                traceback.print_exc()
                raise IOError("When try to save model: {} into disk with error:{}".format(model, e))

    def _check_models(self):
        if not self.model_list:
            raise ValueError("Please provide at least one model! \
                Current model_list is {}".format(len(self.model_list)))


class NeuralNetworkFactory:
    def __init__(self):
        self.neural_networks_name_list = []
    
    @staticmethod
    def get_neural_model_instance(neural_networks_name_list, n_classes):
        """
        n_classes has to be provided as to init instance needs this.
        """
        if isinstance(neural_networks_name_list, str):
            neural_networks_name_list = [neural_networks_name_list]
        
        neural_networks_list = []
        for name in neural_networks_name_list:
            if name == 'DNN':
                neural_networks_list.append(DNNSearch(n_classes))
            elif name == 'CNN':
                pass

        return neural_networks_list


class NeuralModelSearch:
    """
    Main class for caller class to find best models.
    """
    def __init__(self, objective='val_accuracy', 
                    max_trials=5, 
                    executions_per_trial=1, 
                    directory=None, 
                    project_name=None, 
                    algorithm_list=None, 
                    tuning_algorithm='RandomSearch', 
                    num_best_models=5,
                    models_path=None):
        self.objective = objective
        self.max_trials = max_trials
        self.executions_per_trial = executions_per_trial
        self.directory = directory if directory is not None else TMP_FOLDER
        # Here change the project name, as if we use same project, then won't re-fit just `reload`
        # make project name as random string.
        self.project_name = project_name if project_name is not None else self._generate_project_name() 
        
        # based on default hyper yaml file to define which mdoel to use.
        # default_keys = list(hyper_yml.keys())
        # default_keys.remove('optimizers')
        # Add parameter in yaml file will be better
        default_keys = hyper_yml['DefaultAlgorithms']

        self.algorithm_list = default_keys if algorithm_list is None else algorithm_list
        self.tuning_algorithm = tuning_algorithm
        self.num_best_models = num_best_models
        
        # add with models_path for storing the models into path we want.
        self.models_path = models_path

    def fit(self, x, y, epochs=10, val_x=None, val_y=None, validation_split=0.2, evaluate=True):
        # search should be automatically
        
        tuner_list = self._get_search_tuner_list(x, y)

        # whether or not to evaluate should base on attribute
        if evaluate:
            if (not val_x and not val_y):
                train_x, val_x, train_y, val_y = train_test_split(x, y, test_size=validation_split)
        
        # loop for tuner_list to try to search best model
        start_time = time.time()
        for i in range(len(tuner_list)):
            tuner = tuner_list[i]
            logger.info("Start to search neural network models.")
            if not evaluate:
                tuner.search(x, y, epochs=epochs, validation_split=validation_split)
            else:
                tuner.search(train_x, train_y, epochs=epochs, validation_data=(val_x, val_y))

            if evaluate:
                # get each search best models, evaluate and save it.
                best_models = tuner.get_best_models(self.num_best_models)
                
                if not best_models:
                    logger.warning("There is no best model found.")
                    return 
                
                print("Get best models: ", best_models)
                self.evaluate_trained_models(best_models, val_x, val_y)

        # clean folder
        self._clean_search_space()

        logger.info("Whole fitting logic finished used {} seconds.".format(time.time() - start_time))

    def evaluate_trained_models(self, best_models, x, y):
        # get each search best models, evaluate and save it.
        evaluate_model = EvaluateNeuralModel(best_models, x, y)

        best_models_scores = evaluate_model.evaluate_models()
        logger.info("Get best scores are: [{}]".format('\t'.join([str(score) for score in best_models_scores])))

        logger.info("Start to save best trained nueral networks models into disk.")
        evaluate_model.save_models(self.models_path)
    
    def _get_search_tuner_list(self, x, y):
        # first should get class number
        num_classed = get_num_classes_based_on_label(y)
        logger.info("Get {} classes problem.".format(num_classed))

        logger.info("Start to get model instance for algorithms: [{}]".format('\t'.join(self.algorithm_list)))
        self.neural_networks_list = NeuralNetworkFactory.get_neural_model_instance(self.algorithm_list, num_classed)
        
        logger.info("Start to use search algorithm: {} to find models.".format(self.tuning_algorithm))
        
        tuner_list = []
        if self.tuning_algorithm == 'RandomSearch':
            for model in self.neural_networks_list:
                tuner_list.append(RandomSearch(model, 
                    objective=self.objective, 
                    max_trials=self.max_trials, 
                    executions_per_trial=self.executions_per_trial, 
                    directory=self.directory, 
                    project_name=self.project_name))
        else:
            pass
        
        return tuner_list

    @staticmethod
    def _generate_project_name(project_name_size=10, chars=string.ascii_uppercase + string.ascii_lowercase):
        """
        Random generate project name.
        """
        random_project_name = ''.join(random.choice(chars) for _ in range(project_name_size))
        
        return random_project_name

    def _clean_search_space(self):
        """Call only when `fit` logic finished."""
        try:
            logger.info("Try to clean serach model space folder: {}".format(self.project_name))

            search_project_path = os.path.join(self.directory, self.project_name)
            shutil.rmtree(search_project_path)
            logger.info("Folder: {} has been deleted!".format(self.project_name))

        except IOError as e:
            # if we couldn't delete this folder, then OK pass, but should log
            logger.error("To delete search project {} get error: {}".format(self.project_name, e))
            
        

if __name__ == '__main__':
    model = DNNSearch(3)
    tuner = RandomSearch(model, objective='val_accuracy', 
        max_trials=10, executions_per_trial=3, 
        directory='./auto_ml/tmp_folder/tmp',  project_name='test')

    # tuner = NeuralModelRandomSearch(model, objective='val_accuracy', 
    #     max_trials=10, executions_per_trial=3, 
    #     directory='./auto_ml/tmp_folder/tmp',  project_name='test')

    from sklearn.datasets import load_iris
    x, y = load_iris(return_X_y=True)

    # print("Start to search")
    # tuner.search(x, y, epochs=10, validation_split=.2)
    # # tuner.save_best_models(x, y)

    # print("Search step finished.")
    # best_models = tuner.get_best_models(3)

    # # print("Best model score:", [model.evaluate(x, y)[1] for model in best_models])
    # # for model in best_models:
    # #     model_name = "DNN" + str(model.evaluate(x, y)[1])
    # #     model.save(os.path.join(OUTPUT_FOLDER, model_name) + '.h5')

    # evaluate_model = EvaluateNeuralModel(best_models, x, y)
    # print(evaluate_model.evaluate_models())
    # evaluate_model.save_models()

    model_search = NeuralModelSearch()
    model_search.fit(x, y)



# -*- coding:utf-8 -*-
"""
This will contain some model selection logic should be used here like
Grid search logic, as here what I could do is create the search space.
So maybe Grid search to find whole models.
Here is the classifier real training happens here.

@author: Guangqiang.lu
"""
import numpy as np
import time
import tqdm
from sklearn.model_selection import GridSearchCV

from auto_ml.utils.backend_obj import Backend
from auto_ml.utils.logger import create_logger


logger = create_logger(__file__)


class GridSearchModel(object):
    """
    Here could just to implement that could add a list
    of estimators and their parameters list.
    I want to make this class to do real training part.
    """
    def __init__(self, backend, n_best_model=None):
        """
        self.estimator_list is like: [GridSearchCV(lr, params), ...]
        self.score_dict is like: {'LogisticRegressin': (lr, 0.9877)}
        """
        super(GridSearchModel, self).__init__()
        # as we need to do training, so here will just store the trained best model
        # for later step ensemble
        if backend is None:
            raise ValueError("When to use Model Ensemble class, we get a None `backend` object! Please check!")
        self.backend = backend 
        self.estimator_list = []
        self.score_dict = {}
        self.n_best_model = 10 if n_best_model is not None else n_best_model

    def add_estimator(self, estimator, estimator_params=None):
        """
        As I also want to keep current logic with a list
        of estimators in parallel, so here should keep
        whole models.
        As I will my own estimator, so don't need always need
        `estimator_params`, but just add in case we just want
        to add other sklearn estimator
        :param estimator: estimator object
        :param estimator_params: native sklearn object params to search.
        :return:
        """
        if hasattr(estimator, 'get_search_space'):
            # so that this is our estimator
            if self.estimator_list is None:
                self.estimator_list = [GridSearchCV(estimator, estimator.get_search_space())]
            else:
                self.estimator_list.append(GridSearchCV(estimator, estimator.get_search_space()))
        else:
            if estimator_params is None:
                raise ValueError("When we need to set other sklearn native estimator, do"
                                 "need to add estimator_params for searching")

            if self.estimator is None:
                self.estimator_list = [GridSearchCV(estimator, estimator_params)]
            else:
                self.estimator_list.append(GridSearchCV(estimator, estimator_params))

        return self

    def add_estimators(self, estimator_param_pairs):
        """
        This is try to parallel training for whole training
        with different params.
        :param estimator_param_pairs: a list of estimator and params
         just like this:[(lr, {"C":[1, 2]})]
        :return:
        """
        for estimator, estimator_params in estimator_param_pairs:
            self.add_estimator(estimator, estimator_params)

    def fit(self, x, y, n_jobs=None):
        """
        This is real training logic happens here, also we could
        use parallel training for these estimators.
        :param x: training data
        :param y: training label
        :param n_jobs: how much cores to use
        :return:
        """
        # parallel training
        if n_jobs is not None and n_jobs > 1:
            # here couldn't use multiprocessing here, just to set
            # estimator `n_job`
            # we could add other multiprocessing here either if we want.
            logger.info("Start to train model based on {} cores.".format(n_jobs))
            # set `n_jobs` for each estimator
            for estimator in self.estimator_list:
                if hasattr(estimator, 'n_jobs'):
                    estimator.n_jobs = n_jobs

        with tqdm.tqdm(range(len(self.estimator_list))) as process:
            start_time = time.time()
            for i in range(len(self.estimator_list)):
                estimator = self.estimator_list[i]
                estimator.fit(x, y)

                # Add training info, so that we could get better understanding while training happens.
                # Must be `estimator.estimator.name` as estimator is s warpper for each instance.
                logger.info("GridSearch for algorithm: {} takes {} seconds".format(estimator.estimator.name, round(time.time() - start_time, 2)))

                process.update(1)

        logger.info("Model selection training has finished.")

        # after the training finished, then we should get each estimator with score that is based on `training score`
        # and store the score with each instance class name and score.
        self._get_estimators_score()

        # Here add with information for `n_best_model` name and score
        logger.info("Get some best model scores information based on model_selection module.")
        for alg_name, score_tuple in self.score_dict.items():
            logger.info("Algorithm: {} with score: {}".format(alg_name, score_tuple[1]))

        # after we have get the score, then we should store the trained estimators
        logger.info("Start to save best selected models into disk.")
        self.save_best_model_list()

        return self

    def score(self, x, y):
        """
        To use best fitted model to evaluate test data
        :param x:
        :param y:
        :return:
        """
        best_estimator = self.best_estimator

        return best_estimator.score(x, y)

    def predict(self, x):
        """
        Get prediction based on best fitted model
        :param x:
        :return:
        """
        best_estimator = self.best_estimator

        if hasattr(best_estimator, 'predict'):
            return best_estimator.predict(x)
        else:
            raise NotImplementedError("For estimator:{} doesn't support `predict` func!".format(best_estimator))

    def predict_proba(self, x):
        """
        Get probability of based on best estimator
        :param x:
        :return:
        """
        best_estimator = self.best_estimator

        if hasattr(best_estimator, 'predict_proba'):
            return best_estimator.predict_proba(x)
        else:
            raise NotImplementedError("For estimator:{} doesn't support `predict_proba` func!".format(best_estimator))

    def save_best_model_list(self):
        """
        save whole best fitted model based on each algorithm own parameters, so
        that we could save each best models.
        Then we could do ensemble logic.

        Here I think I could just save the each best parameters trained model
        into disk, also the file name should be like `LogisticRegression_9813.pkl`:
        with `classname_score.pkl`.

        Noted: This func should only be called after trained
        :param n_best_model: How many best model to save
        :return:
        """
        # added how many best model to be saved into disk based on score.
        n_best_score_dict = {k: v for k, v in sorted(self.score_dict.items(),
                                                     key=lambda x: x[1][1], reverse=True)[:self.n_best_model]}

        for estimator_name, estimator_tuple in n_best_score_dict.items():
            estimator = estimator_tuple[0]
            # NO NEED: just save the score float into str
            # Get score string with round 6 or 100% accuracy
            # score_str = str(round(estimator_tuple[1], 6)).split('.')[-1] \
            #     if estimator_tuple[1] < 1.0 else '100'
            score_str = str(round(estimator_tuple[1], 6))

            # ADD with `-` with name and score, so that we could get the score later
            model_name = estimator_name + "-" + score_str
            logger.info("Start to save model: {}".format(model_name))
            self.backend.save_model(estimator, model_name)

        logger.info("Already have saved models: %s" % '\t'.join(self.backend.list_models()))

    def load_best_model_list(self, model_extension='pkl'):
        """
        Load previous saved best model into a list of trained instance.
        :return:
        """
        model_list = []
        n_best_score_dict = {k: v for k, v in sorted(self.score_dict.items(),
                                                     key=lambda x: x[1][1], reverse=True)[:self.n_best_model]}

        for estimator_name, estimator_tuple in n_best_score_dict.items():
            # TODO: this is used for `sklearn`module, how about if we need other framework?
            model_name = estimator_name + "-" + str(round(estimator_tuple[1], 6)) + model_extension
            try:
                model_instance = self.backend.load_model(model_name)
                model_list.append(model_instance)
            except IOError as e:
                raise IOError("When try load trained model file:{} "
                              "with error: {}".format(estimator_name, e))
        return model_list

    def save_bestest_model(self):
        """
        dump best trained model into disk, one more thing here: We shouldn't save
        the model into disk with fixed name, but should with score.
        :return:
        """
        self.best_esmator_name = self.best_estimator.__class__.__name__ + '-' + str(self.best_score) + ".pkl"
        self.backend.save_model(self.best_estimator,  self.best_esmator_name)

    def load_bestest_model(self):
        """
        load best trained model from disk
        :return:
        """
        return self.backend.load_model(self.best_esmator_name)

    @property
    def best_estimator(self):
        """
        To get best estimator based on the training score list
        :return: best estimator
        """
        best_estimator_list = [estimator.best_estimator_ for estimator in self.estimator_list]
        best_score_list = [estimator.best_score_ for estimator in self.estimator_list]

        if best_estimator_list is not None:
            if best_score_list is None:
                logger.error("We to get best estimator, don't get best score list")
                raise Exception("We to get best estimator, don't get best score list")

            best_estimator = best_estimator_list[np.argmax(best_score_list)]

            return best_estimator
        else:
            logger.error("When to get best estimator, we don't get the best estimator list")
            raise Exception("When to get best estimator, we don't get the best estimator list")

    @property
    def best_score(self):
        best_score_list = [estimator.best_score_ for estimator in self.estimator_list]
        try:
            return round(max(best_score_list), 6)
        except Exception as e:
            raise Exception("When try to get best score get error: {}".format(e))

    def _get_estimators_score(self):
        """
        To get whole trained estimator based on data and label for storing
        the result based on each trained grid model best estimator.
        score_dict is like: {'LogisticRegressin': (lr, 0.9877)}
        :return:
        """
        for estimator in self.estimator_list:
            # here I also need the trained estimator object, so here
            # also with trained object.
            best_estimator = estimator.best_estimator_
            best_score = round(estimator.best_score_, 6)

            # This should based on the trained best score, not based on trained model then score again.
            self.score_dict[best_estimator.__class__.__name__] = (best_estimator, best_score)
                                               #self._score_with_estimator(best_estimator, x, y))

    @staticmethod
    def _score_with_estimator(estimator_instance, x, y):
        """
        Just to get score with trained estimator based on data and label
        :param estimator_instance:
        :param x:
        :param y:
        :return:
        """
        try:
            score = estimator_instance.score(x, y)
            return score
        except Exception as e:
            raise Exception("When try to get score with estimator: {} "
                            "get error: {}".format(estimator_instance.__class__.__name__, e))

    def save_trained_estimator(self, estimator, estimator_name):
        """
        To save the trained model into disk with file name
        :param estimator:
        :param estimator_name:
        :return:
        """
        self.backend.save_model(estimator, estimator_name)

    def print_estimators(self):
        """
        To list whole grid models instances, so that we could check.
        :return:
        """
        logger.info("Get {} estimators.".format(len(self.estimator_list)))
        for grid_estimator in self.estimator_list:
            print(grid_estimator)


if __name__ == '__main__':
    from sklearn.datasets import load_iris
    from auto_ml.base.classifier_algorithms import LogisticRegression
    from auto_ml.base.classifier_algorithms import SupportVectorMachine
    from auto_ml.utils.backend_obj import Backend

    backend = Backend()

    x, y = load_iris(return_X_y=True)

    g = GridSearchModel(backend)
    lr = LogisticRegression()
    clf = SupportVectorMachine()
    g.add_estimator(lr)
    g.add_estimator(clf)

    g.fit(x, y)
    print(g.best_estimator)
    print(g.best_score)
    print(g.score(x, y))

    g.save_bestest_model()
    bst_model = g.load_bestest_model()
    print(bst_model.score(x, y))
    print(g.score_dict)
    g.save_best_model_list()
    print(g.load_best_model_list())



# -*- coding:utf-8 -*-
"""
this is to write a class that we could use for
whole algorithms that we used in this class.

@author: Guangqiang.lu
"""
import numpy as np
import os

from auto_ml.utils.paths import load_param_config


class ConfigSpace(object):
    def __init__(self):
        self.config = load_param_config()
        self.algorithm_config = None

    def load_sklearn_params(self, algorithm_name):
        if algorithm_name == 'LogisticRegression':
            self.algorithm_config = self._load_lr_params()
        else:
            pass

    def _load_lr_params(self):
        config_lr = self.config['sklearn']['classification']['LogisticRegression']
        config_lr['C'] = np.random.exponential(1, size=5)
        config_lr['fit_intercept'] = [True, False]
        config_lr['penalty'] = ['l1', 'l2']

        return config_lr


if __name__ == '__main__':
    config = ConfigSpace()
    config.load_sklearn_params('LogisticRegression')
    lr_config = config.algorithm_config

    from auto_ml.base.classifier_algorithms import LogisticRegression
    from sklearn.model_selection import GridSearchCV
    from sklearn.datasets import load_iris
    x, y = load_iris(return_X_y=True)

    lr = LogisticRegression()
    estimator = GridSearchCV(lr, lr_config, cv=10)
    estimator.fit(x, y)
    print(estimator.best_score_)



# -*- coding:utf-8 -*-
from tensorflow.keras.layers import Input, Dense, Dropout, BatchNormalization, Activation
from tensorflow.keras.models import Model

from auto_ml.nueral_networks.basic_net import BaseNet


class DnnNet(BaseNet):
    """This is basic DNN networks.
        Arguments:
            n_classes(Integer): How many classes you want to classifier or use it for regression.
                Integer 2 is for binary problem, Integer >2 is for how many classes you have,
                Integert -1 is for regression problem.

            n_dims(Integer): How many dimensions of training data.

            n_layers(Integer, optional): How many dense layers wanted to be used.

            n_units(Integer, optional): How many units to be used for each dense layer.

            use_dropout(Boolean, optional): Whether or not to use Dropout after each layer.
                Default is True, recommended when model is overfitting.

            drop_ratio(Double, optional): How many ratios to drop units during training.
                Default is 0.5, if wanted for bigger penalty, set this bigger, may cause underfitting.

            use_batch(Boolean, optional): Whether or not to use BatchNormalization durng training.
                Default is True, this is great for improving model training accuracy

            loss(String, optional): Which loss function wanted to be used during training as objective function.
                Default is None, it will be choosen according to different problems. For 'binary' problem, use
                'binary_crossentropy', for 'multi-class' problem, use 'categorical_crossentropy', for 'regression'
                problem, use 'mse'.

            metrics(String, optional): Which metric wanted to be used during training.

            optimizer(String, optional): Which optimizer to be used for optimizing.

            silence(Boolean, optional): Whether to print some useful information to console.
                Default is False, means print something to console.
    """
    def __init__(self, n_classes, n_dims, n_layers=3, n_units=64, use_dropout=True, drop_ratio=.5,
                 use_batch=True, loss=None, metrics='accuracy', optimizer='adam', silence=False):
        super().__init__()
        self.n_classes = n_classes
        self.n_dims = n_dims
        self.n_layers = n_layers
        self.n_units = n_units
        self.use_dropout = use_dropout
        self.drop_ratio = drop_ratio
        self.use_batch = use_batch
        self.loss = loss
        self.metrics = metrics
        self.optimizer = optimizer
        self.silence = silence
        self.model = self._init_model()

    # this is dense block function.
    def _dense_block(self, layers):
        res = Dense(self.n_units)(layers)
        if self.use_batch:
            res = BatchNormalization()(res)
        res = Activation('relu')(res)
        if self.use_dropout:
            res = Dropout(self.drop_ratio)(res)
        return res

    def _init_model(self):
        inputs = Input(shape=(self.n_dims, ))
        for i in range(self.n_layers):
            if i == 0:
                res = self._dense_block(inputs)
            else:
                res = self._dense_block(res)

        # this is method private function to check whether or not loss is not given, then use default loss
        def _check_loss(model, loss, metrics, optimizer):
            if loss is not None:
                model.compile(loss=loss, metrics=[metrics], optimizer=optimizer)
            return model

        if self.n_classes == 2:      # this is binary class problem.
            out = Dense(self.n_classes, activation='sigmoid')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='binary_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes >= 2:    # this is multiclass problem.
            out = Dense(self.n_classes, activation='softmax')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='categorical_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else: _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes == -1:     # this is regression problem
            out = Dense(1)(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='mse', metrics=[self.metrics], optimizer=self.optimizer)
            else: _check_loss(model, self.loss, self.metrics, self.optimizer)
        else:
            raise AttributeError("Parameter 'n_classes' should be -1, 2 or up 2!")

        print('Model structure summary:')
        model.summary()

        return model


if __name__ == '__main__':
    from auto_ml.utils.logger import create_logger

    from sklearn.datasets import load_iris

    logger = create_logger()

    x, y = load_iris(return_X_y=True)



    dnn = DnnNet(3, 4, n_layers=1, n_units=128)
    dnn.fit(x, y)

    logger.info(dnn.score(x, y))
    logger.info("Sample prob:", dnn.predict_proba(x)[:5])
    logger.info("Sample pred:", dnn.predict(x)[:5])

    



# -*- coding:utf-8 -*-
"""
To process data with standard logic

@author: Guangqiang.lu
"""
from sklearn.preprocessing import StandardScaler
from auto_ml.preprocessing.processing_base import Process


class Standard(Process):
    def __init__(self):
        super(Standard, self).__init__()
        self.estimator = StandardScaler()


if __name__ == '__main__':
    from sklearn.datasets import load_iris
    x, y = load_iris(return_X_y=True)

    s = Standard()
    s.fit(x, y)
    print(s.transform(x))
    print(s.fit_transform(x))



# -*- coding:utf-8 -*-
"""
This class is used as a `factory` to get instance that we need for the whole project using
`factory pattern`.

@author: Guangqiang.lu
"""
from auto_ml.preprocessing import \
    (onehotencoding,  standardization,  norlization,  minmax,  imputation,  feature_selection,  pca_reduction)


class ProcessingFactory:
    @staticmethod
    def get_processor_list(processor_name_list):
        """
        Simple factory.
        
        As sklearn pipeline is a list of tuple(name,  transform, so here will suit for it.
        :param processor_name_list: which algorithms to include.
        :return:
        """
        processor_tuple = []

        if isinstance(processor_name_list,  str):
            processor_name_list = [processor_name_list]

        for processor_name in processor_name_list:
            if processor_name == 'Imputation':
                processor_tuple.append((processor_name, imputation.Imputation()))
            elif processor_name == 'OnehotEncoding':
                processor_tuple.append((processor_name, onehotencoding.OnehotEncoding()))
            elif processor_name == 'Standard':
                processor_tuple.append((processor_name, standardization.Standard()))
            elif processor_name == 'Normalize':
                processor_tuple.append((processor_name, norlization.Normalize()))
            elif processor_name == 'MinMax':
                processor_tuple.append((processor_name, minmax.MinMax()))
            elif processor_name == 'FeatureSelect':
                processor_tuple.append((processor_name, feature_selection.FeatureSelect()))
            elif processor_name == 'PrincipalComponentAnalysis':
                processor_tuple.append((processor_name, pca_reduction.PrincipalComponentAnalysis()))

        return processor_tuple



# -*- coding:utf-8 -*-
"""
This is used for whole classification problem that could be used
with sklearn backend engine.
author: Guangqiang.lu
"""
from auto_ml.estimator import ClassificationEstimator

if __name__ == '__main__':
    # this is unittest for classification functionality.
    from sklearn.datasets import load_iris
    from sklearn.model_selection import train_test_split

    x, y = load_iris(return_X_y=True)
    xtrain, xtest, ytrain, ytest = train_test_split(x, y, test_size=.2, random_state=123)

    estimator = ClassificationEstimator()

    estimator.fit(xtrain, ytrain, xtest, ytest)

    print(estimator.score(xtest, ytest))





# -*- coding:utf-8 -*-
"""
This is whole path functionality that we could use
in this project.

@author: Guangqiang.lu
"""
import os
from pathlib import Path
import yaml

from auto_ml.utils.func_utils import deprecated


def get_root_path():
    """
    this is to get the root path of the code
    :return: path
    """
    path = str(Path(__file__).parent.parent)
    return path


def get_param_config_path():
    """
    this is to get the param config path
    :return: config file path
    """
    path = str(Path(__file__).parent)
    config_path = os.path.join(path,
                               [x for x in os.listdir(path)
                                if x == 'param_config.yml'][0])
    return config_path


@deprecated
def load_param_config():
    """
    this is to load yaml config object.
    :return: dictionary config
    """
    try:
        with open(get_param_config_path(), 'r') as f:
            config = yaml.safe_load(f)

        return config
    except yaml.YAMLError as e:
        raise IOError("When try to read config file with error: %s" % e)


def load_yaml_file(file_name=None, file_path=None):
    """
    To load yaml file from server path.
    :param file_path: where the file exist
    :return: dictionary
    """
    try:
        if file_name is None:
            # if we don't provide the file_path, try to load root path with file_name: default_algorithms.yaml
            file_name = 'default_algorithms.yml'

        if not file_path:
            file_path = os.path.join(get_root_path(), file_name)
        else:
            file_path = os.path.join(file_path, file_name)

        with open(file_path, 'r') as f:
            data = yaml.safe_load(f)
        return data
    except yaml.YAMLError as e:
        raise IOError("When try to load yaml file from path: {} get error: {}".format(file_path, e))


def get_file_base_name(path):
    """
    To get base file name based on the path
    :param path:
    :return:
    """
    try:
        name = os.path.basename(path).split('.')[0]
        return name
    except ValueError as e:
        raise ValueError("When to get file path: {} with error: {}".format(path, e))


if __name__ == "__main__":
    # print(get_param_config_path())

    print(get_root_path())



# -*- coding:utf-8 -*-
"""
This is to normalize the data to a normal distribution for some algorithms

@author: Guangqiang.lu
"""
from sklearn.preprocessing import Normalizer
from auto_ml.preprocessing.processing_base import Process


class Normalize(Process):
    def __init__(self):
        super(Normalize, self).__init__()
        self.estimator = Normalizer()


if __name__ == '__main__':
    from sklearn.datasets import load_iris
    x, y = load_iris(return_X_y=True)

    norm = Normalize()
    norm.fit(x)
    print(norm.transform(x))
    print(norm.name)



# -*- coding:utf-8 -*-
"""
Let's just use whole regression used in sklearn to be instant
here so that we could change or do something change could be easier.

@author: Guangqiang.lu
"""
from auto_ml.hyper_config import (ConfigSpace, UniformHyperparameter, CategoryHyperparameter,
                                                          GridHyperparameter)
from sklearn.base import BaseEstimator


class RegressorClass(BaseEstimator):
    def fit(self, x, y):
        raise NotImplementedError

    def predict(self, x):
        pred = self.estimator.predict(x)
        return pred

    def score(self, x, y):
        score = self.estimator.score(x, y)
        return score

    @staticmethod
    def get_search_space():
        raise NotImplementedError


class LinearRegression(RegressorClass):
    def __init__(self, fit_intercept=True,
                 n_jobs=None,
                 normalize=False
                 ):
        self.fit_intercept = fit_intercept
        self.n_jobs = n_jobs
        self.normalize = normalize
        self.estimator = None

    def fit(self, x, y):
        from sklearn.linear_model import LinearRegression

        self.estimator = LinearRegression(fit_intercept=self.fit_intercept,
                                          n_jobs=self.n_jobs,
                                          normalize=self.normalize)

        self.estimator.fit(x, y)

        return self

    def predict(self, x):
        return super().predict(x)

    def score(self, x, y):
        return super().score(x, y)


# -*- coding:utf-8 -*-
"""
This is the whole score functionality that could be used for metrics.

author: Guangqiang.lu
"""

from abc import ABCMeta
import numpy as np
from sklearn import metrics
from sklearn.utils.multiclass import type_of_target


class Scorer(object, metaclass=ABCMeta):
    """
    This class is used to make the scorer function to instance
    scorer object
    """
    def __init__(self, name, score_func, max_or_min, sign, kwargs):
        """
        init scorer object
        :param name: which name the metric is
        :param score_func: which scorer function to use
        :param max_or_min: as for different problem, as for classification
        problem best score is 1, and for regression problem min score is 0
        :param sign: the sign object for metrics, like -1 or 1
        :param kwargs: other key words
        """
        self.name = name
        self._score_func = score_func
        self._max_or_min = max_or_min
        self._sign = sign
        self._kwargs = kwargs

    @classmethod
    def __call__(self, y_true, y_pred, sample_weights=None, *args, **kwargs):
        """
        This is the abstract function that subclass should implement
        :param y_true: True label array
        :param y_pred: prediction label array
        :param sample_weights: each sample weights, default is None
        :param args: other keys
        :param kwargs: other key words
        :return: score computed with self._score_func
        """
        pass

    def __str__(self):
        return self.name


# Here is to make the metric class for supporting prediction evaluation,
# probability evaluation and threshould evaluation
class _PredScorer(Scorer):
    def __call__(self, y_true, y_pred, sample_weight=None, *args, **kwargs):
        """
        implement the Scorer caller function, just support for classification,
        not support for regression
        :param y_true: array-like, [n_sampels,] or [n_samples * n_classes]
            true label array
        :param y_pred: array-like, [n_samples,] or [n_samples * n_classes]
            prediction label array
        :param sample_weight: array-like, optional, default is None
            each sample weights, default is None
        :param args:
            other keys
        :param kwargs:
            other key words
        :return:
            computed score with prediction
        """
        true_type = type_of_target(y_true)
        if len(y_true.shape) == 1 and true_type == 'continous':
            # not used for regression problem
            pass
        elif true_type == 'binary' or true_type == 'multiclass':
            if len(y_pred.shape) != 1:
                # if this is one-hot type object
                y_pred = np.argmax(y_pred, axis=1)
        elif true_type == 'multilabel-indicator':
            # here is multi-label classification
            y_pred[y_pred > .5] = 1.
            y_pred[y_pred <= .5] = 0.

        if sample_weight is not None:
            # if we could multiply each sample prediction with weights
            return self._sign * self._score_func(y_true, y_pred,
                                                 sample_weight=sample_weight,
                                                 **kwargs)
        else:
            return self._sign * self._score_func(y_true, y_pred,
                                                 **kwargs)


class _ProbScorer(Scorer):
    def __call__(self, y_true, y_pred, sample_weight=None, *args, **kwargs):
        """
        this class is used to compute the score with probability based,
        y_pred should be the probability!
        :param y_true: array-like, [n_samples * n_classes]
            true label array
        :param y_pred: array-like, [n_samples * n_classes]
            probability array
        :param sample_weight: array-like, optional, default is None
            weights for each sample
        :param args:
            other keys
        :param kwargs:
            other key words
        :return:
            computed score with probability
        """
        true_type = type_of_target(y_true)
        if len(y_true.shape) == 0 or true_type == 'continous':
            pass

        if sample_weight is not None:
            return self._sign * self._score_func(y_true, y_pred,
                                                 sample_weight=sample_weight,
                                                 **kwargs)
        else:
            return self._sign * self._score_func(y_true, y_pred,
                                                 **kwargs)


class _ThreaScorer(Scorer):
    def __call__(self, y_true, y_pred, sample_weight=None, *args, **kwargs):
        """
        this class is used to compute the score with probability based,
        y_pred should be the probability!
        :param y_true: array-like, [n_samples * n_classes]
            true label array
        :param y_pred: array-like, [n_samples * n_classes]
            probability array
        :param sample_weight: array-like, optional, default is None
            weights for each sample
        :param args:
            other keys
        :param kwargs:
            other key words
        :return:
            computed score with probability
        """
        y_type = type_of_target(y_true)
        if y_type not in ('binary', 'multilabel-indicator'):
            raise ValueError("{} format is not supported".format(y_type))

        if y_type == 'binary':
            # we chould just get the last column probability
            y_pred = y_pred[:, 1]
        elif y_type == 'multilabel-indicator':
            y_pred = np.vstack([p[:, -1] for p in y_pred]).T

        if sample_weight is not None:
            return self._sign * self._score_func(y_true, y_pred,
                                                 sample_weight=sample_weight,
                                                 **kwargs)
        else:
            return self._sign * self._score_func(y_true, y_pred,
                                                 **kwargs)


# here is a helper function to make scorer for later step use case
def make_scorer(name, score_func, max_or_min=1, greater_is_better=True,
                need_proba=False, need_thre=False, **kwargs):
    """
    This is to make scorer object just like sklearn make_scorer
    :param name: string-type
        the name for the function
    :param score_func: callable function object
        as sklearn score object, `score_func(y_true, y_pred, **kwargs)`
    :param max_or_min: int or float
        the max score for accuracy and min score for loss function, that means
        if score function is accuracy, then this should be 1, if score function
        is loss function, then this should be 0
    :param greater_is_better: Boolean, True or False
        for different metrics, whether bigger is better, like for accuracy like
        metrics, bigger is better, but for loss function, smaller is better
    :param need_proba: Boolean
        whether or not to use the probability, sometimes like AUC score is based on
        probability, then we should give the proba that metrics need
    :param need_thre: Boolean
        whether or not to use threshold to compute the score like AUC score
    :param kwargs:
        other key words
    :return:
        metrics object
    """
    sign = 1 if greater_is_better else -1
    if need_proba:
        cls = _ProbScorer
    if need_thre:
        cls = _ThreaScorer
    else:
        cls = _PredScorer
    return cls(name, score_func, max_or_min, sign, kwargs)


# Here should create with standard regression metrics
r2 = make_scorer('r2',
                 metrics.r2_score)

mean_absolute_error = make_scorer('mean_absolute_error',
                                  metrics.mean_absolute_error,
                                  max_or_min=0,
                                  greater_is_better=False)

mean_squared_error = make_scorer('mean_squared_error',
                                 metrics.mean_squared_error,
                                 max_or_min=0,
                                 greater_is_better=False)


# here is standard classification problem metrics
accuracy = make_scorer('accuracy', metrics.accuracy_score)

f1 = make_scorer('f1', metrics.f1_score)

roc_auc = make_scorer('roc_auc', metrics.roc_auc_score, need_proba=True)

precision = make_scorer('precision', metrics.precision_score)

recall = make_scorer('recall', metrics.recall_score)

average_precision = make_scorer('average_precision', metrics.average_precision_score)




# -*- coding:utf-8 -*-
"""
This is whole constant value that we would use in the
whole project, as like what kind of problem that we
support, current we just support classification, regression.

@author: Guangqiang.lu
"""
BINARY_CLASSIFICATION = 0
MULTI_CLASS_CLASSIFICATION = 1
MULTI_LABEL_CLASSIFICATION = 2
REGRESSION = 3
CLUSTERING = 4
RECOMMENDATION = 5

# Here should confirm that there are 4 type tasks are supported
CLASSIFICTION_TASK = [BINARY_CLASSIFICATION, MULTI_CLASS_CLASSIFICATION, MULTI_LABEL_CLASSIFICATION]
REGRESSION_TASK = [REGRESSION]
CLUSTERING_TASK = [CLUSTERING]
RECOMMENDATION_TASK = [RECOMMENDATION]

WHOLE_TASK = CLASSIFICTION_TASK + REGRESSION_TASK + CLUSTERING_TASK + RECOMMENDATION_TASK

# Here make a dictionary to for recognize which task it is
TASK_TO_STRING = {BINARY_CLASSIFICATION: "binary",
                  MULTI_CLASS_CLASSIFICATION: "multiclass",
                  MULTI_LABEL_CLASSIFICATION: "multilabel",
                  REGRESSION: "regression",
                  CLUSTERING: "clustering",
                  RECOMMENDATION: "recommendation"}

# Here is to make string to task names
STRING_TO_TASK = {v: k for k, v in TASK_TO_STRING.items()}

# Here I just add a tmp folder path and model save path
import tempfile
import os
from pathlib import Path

# Added logic to keep only one temp folder is used no other folders are created for this project.
# TODO:Have to write the folder names into disk, otherwise how we know the only one tmp folder?
def get_tmp_folder():
    cur_path = os.path.abspath(Path(__file__).parent)

    print("Get cur_path: ", cur_path)
    try:
        config_file_path = os.path.join(cur_path, 'config.txt')
        if not os.path.exists(config_file_path):
            return tempfile.mkdtemp() 

        with open(config_file_path, 'r') as f:
            data = f.read()

        # with read couldn't just is None, not None in fact
        if len(data) == 0:
            with open(config_file_path, 'w') as f:
                PROJECT_TMP_PATH = tempfile.mkdtemp()
                f.write(PROJECT_TMP_PATH)
        else:
            PROJECT_TMP_PATH = data

        return PROJECT_TMP_PATH
    except IOError as e:
        print("When try to read and write `config.txt` file with error: {}".format(e))
   
PROJECT_TMP_PATH = get_tmp_folder()
print("******* Get temp folder *******: {}".format(PROJECT_TMP_PATH)) 

# :TODO: in real prod, change this.
TMP_FOLDER = os.path.join(PROJECT_TMP_PATH, "tmp")
OUTPUT_FOLDER = os.path.join(PROJECT_TMP_PATH, "models")

# Add validation split threshold
VALIDATION_THRESHOLD = 10000


# -*- coding:utf-8 -*-
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Dropout, BatchNormalization, Activation, Conv2D, Conv1D, \
     GlobalAveragePooling1D, GlobalAveragePooling2D, concatenate, Flatten, add

from auto_ml.nueral_networks.basic_net import BaseNet


class WideDeepNet(BaseNet):
    """This class implements Wide & Deep deep learning models.

            Arguments:
                n_classes(Integer): How many classes you want to classifier or use it for regression.
                    Integer 2 is for binary problem, Integer >2 is for how many classes you have,
                    Integert -1 is for regression problem.

                input_dim_1(Integer): First Dimension of your data sets. Noted, this is not batch-size dimension.

                input_dim_2(Integer): Second Dimension of your data sets.

                input_dim_3(Integer, optional): Third Dimension of your data sets. Noted, if data sets is just 2D,
                    this parameter should not be given.

                is_2D(Boolean, optional): Whether or not data sets is just 2D.
                    Default is True.

                is_3D(Boolean, optional): Whether or not data sets is just 3D.
                    Default is False.

                n_wide_layers(Integer, optional): How many wide layers to be used.

                first_conv(Boolean, optional): Whether to use Convolutional layer at before layers.

                n_first_convs(Integer, optional): How many Convolutional layers to be used at before layers.

                conv_units(Integet, optional): How many units for each Convolutional layer.
                    For better accuracy, give more units will be good.

                kernel_size(Integet, optional): How big is kernel wanted to slide on data sets.
                    This kernel will be (n*n)

                strides(Integer, optional): How many steps make the kernel slide on data sets.

                padding(String, optional): Whether or not drop or fill the missing values.
                    Default is 'SAME': returned result is same as before, if 'valid': missing value will be droped.

                use_global(Boolean, optional): Whether or not to use global average pooling.
                    Default is False, meaning Flatten Conv result.

                use_pooling(Boolean, optional): Whether or not to use pooling after Convolutional layer.
                    Default if False, meaning without Pooling layer.

                use_dropout(Boolean, optional): Whether or not to use Dropout after each layer.
                    Default is True, recommended when model is overfitting.

                drop_ratio(Double, optional): How many ratios to drop units during training.
                    Default is 0.5, if wanted for bigger penalty, set this bigger, may cause underfitting.

                use_batch(Boolean, optional): Whether or not to use BatchNormalization durng training.
                    Default is True, this is great for improving model training accuracy

                activation(String, optional): Which activation function wanted to be used.
                    Default is 'relu'

                use_dnn(Boolean, optional): Whether or not to use the Dense layer.
                    Default is True

                n_dnn_layers(Integer, optional): How many dense layer wanted to be used.

                dnn_units(Integer, optional): How many units to be used for each layer.

                loss(String, optional): Which loss function wanted to be used during training as objective function.
                    Default is None, it will be choosen according to different problems. For 'binary' problem, use
                    'binary_crossentropy', for 'multi-class' problem, use 'categorical_crossentropy', for 'regression'
                    problem, use 'mse'.

                metrics(String, optional): Which metric wanted to be used during training.

                optimizer(String, optional): Which optimizer to be used for optimizing.

                silence(Boolean, optional): Whether to print some useful information to console.
                    Default is False, means print something to console.

        """
    def __init__(self, n_classes, input_dim_1, input_dim_2, input_dim_3=None, is_2D=True, is_3D=False,
                 n_wide_layers=3, first_conv=True, n_first_convs=1, layer_concat=False, kernel_size=2, stridess=1, padding='SAME',
                 use_global=False, use_pooling=False, use_dnn=True, n_dnn_layers=1, conv_units=64, dnn_units=128, use_dropout=True,
                 drop_ratio=.5, use_batch=True, activation='relu', loss=None, optimizer='rmsprop', metrics='accuracy', silence=False):
        super().__init__()
        self.n_classes = n_classes
        self.input_dim_1 = input_dim_1
        self.input_dim_2 = input_dim_2
        self.input_dim_3 = input_dim_3
        self.is_2D = is_2D
        self.is_3D = is_3D
        self.n_classes = n_classes
        self.n_wide_layers = n_wide_layers
        self.first_conv = first_conv
        self.n_first_convs = n_first_convs
        self.layer_concat = layer_concat
        self.use_global = use_global
        self.use_dnn = use_dnn
        self.use_pooling = use_pooling
        self.n_dnn_layers = n_dnn_layers
        self.conv_units = conv_units
        self.kernel_size = kernel_size
        self.strides = stridess
        self.padding = padding
        self.dnn_units = dnn_units
        self.use_dropout = use_dropout
        self.drop_ratio = drop_ratio
        self.use_batch = use_batch
        self.activation = activation
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
        self.silence = silence
        self.model = self._init_model()

    # This private function is to construct wide deep block.
    def _wide_conv(self, conv, layers):
        wide_1 = conv(self.conv_units, 1, padding=self.padding)(layers)

        wide_2 = conv(self.conv_units, 1, padding=self.padding)(layers)
        wide_2 = conv(self.conv_units, 3, padding=self.padding)(wide_2)

        wide_3 = conv(self.conv_units, 1, padding=self.padding)(layers)
        wide_3 = conv(self.conv_units, 5, padding=self.padding)(wide_3)

        wide_4 = conv(self.conv_units, 3, padding=self.padding)(layers)
        wide_4 = conv(self.conv_units, 1, padding=self.padding)(wide_4)

        # whether or not to use conconcate to merge layer or just add them up.
        if self.layer_concat:
            return concatenate([wide_1, wide_2, wide_3, wide_4], axis=0)
        else:
            return add([wide_1, wide_2, wide_3, wide_4])

    # This is wide and deep block construction function.
    # GoogleNet Inception block
    def _wide_deep_block(self, layers):
        if self._check_dims(self.input_dim_3, self.is_2D, self.is_3D) == '2D':
            res = self._wide_conv(Conv1D, layers)
        elif self._check_dims(self.input_dim_3, self.is_2D, self.is_3D) == '3D':
            res = self._wide_conv(Conv2D, layers)

        return res

    # Main model sturcture construction function
    def _init_model(self):
        # Check given parameters to judge data is '2D' or '3D'
        self.dimstr = self._check_dims(self.input_dim_3, self.is_2D, self.is_3D)
        
        if self.dimstr == '2D':
            inputs = Input(shape=(self.input_dim_1, self.input_dim_2))
        elif self.dimstr == '3D':
            inputs = Input(shape=(self.input_dim_1, self.input_dim_2, self.input_dim_3))

        # Whether first use Convolutional layer, can be choosen.
        if self.first_conv:
            for _ in range(self.n_first_convs):
                if self.dimstr == '2D':
                    res_conv = Conv1D(self.conv_units, self.kernel_size, self.strides, padding=self.padding)(inputs)
                elif self.dimstr == '3D':
                    res_conv = Conv2D(self.conv_units, self.kernel_size, self.strides, padding=self.padding)(inputs)
        else:
            res_conv = inputs

        # Here is Wide & Deep model Block
        for i in range(self.n_wide_layers):
            if i == 0:
                res = self._wide_deep_block(res_conv)
            else:
                res = self._wide_deep_block(res)

        # Whether to use global avarega pooling or just Flatten concolutional result
        if self.use_global:
            if self.dimstr == '2D':
                res = GlobalAveragePooling1D()(res)
            elif self.dimstr == '3D':
                res = GlobalAveragePooling2D()(res)
        else:
            res = Flatten()(res)

        # Whether to use Dense layers
        if self.use_dnn:
            for _ in range(self.n_dnn_layers):
                res = Dense(self.dnn_units)(res)
                if self.use_batch:
                    res = BatchNormalization()(res)
                res = Activation(self.activation)(res)
                if self.use_dropout:
                    res = Dropout(self.drop_ratio)(res)

        # this is method private function to check whether or not loss is not given, then use default loss
        def _check_loss(model, loss, metrics, optimizer):
            if loss is not None:
                model.compile(loss=loss, metrics=[metrics], optimizer=optimizer)
            return model

        if self.n_classes == 2:      # this is binary class problem.
            out = Dense(self.n_classes, activation='sigmoid')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='binary_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:  _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes >= 2:    # this is multiclass problem.
            out = Dense(self.n_classes, activation='softmax')(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='sparse_categorical_crossentropy', metrics=[self.metrics], optimizer=self.optimizer)
            else:
                _check_loss(model, self.loss, self.metrics, self.optimizer)
        elif self.n_classes == -1:     # this is regression problem
            out = Dense(1)(res)
            model = Model(inputs, out)
            if self.loss is None:
                model.compile(loss='mse', metrics=[self.metrics], optimizer=self.optimizer)
            else: _check_loss(model, self.loss, self.metrics, self.optimizer)
        else:
            raise AttributeError("Parameter 'n_classes' should be -1, 2 or up 2!")

        print('Model structure summary:')
        model.summary()

        return model

    @staticmethod
    def _check_dims(dims, is_2d, is_3d):
        if is_2d and dims is None:
            return '2D'
        elif is_3d or dims is not None:
            return '3D'


if __name__ == '__main__':
    import numpy as np

    data = np.random.randn(100, 10, 10)
    label = np.array([np.random.randint(0, 3) for _ in range(100)]).reshape(-1, 1)
    wide_deep_model = WideDeepNet(3, 10, 10)

    his = wide_deep_model.fit(data, label)

    score = wide_deep_model.score(data, label)

    model_name = wide_deep_model.__class__.__name__ + '-' + str(score)

    wide_deep_model.save(model_name)

    prob = wide_deep_model.predict_proba(data)
    pred = wide_deep_model.predict(data)

    print("Sample prob:", prob[:3])
    print("Sample pred:", pred[:3])




# -*- coding:utf-8 -*-
"""
First step should cover missing field logic.
Here contain both algorithm based filling and frequency filling,
also if one columns contain too many missing data, then will
remove that column.

@author: Guangqiang.lu
"""
import numpy as np
import pandas as pd
from sklearn.impute import KNNImputer, SimpleImputer
from auto_ml.preprocessing.processing_base import Process
from auto_ml.utils.data_rela import is_categorical_type


class Imputation(Process):
    def __init__(self, use_al_to_im=False, threshold=.5):
        super(Imputation, self).__init__()
        self.use_al_to_im = use_al_to_im
        self.threshold = threshold

    def fit(self, data, y=None):
        """
        Imputation logic happens here.
        I have to add a logic to process with different type of columns,
        like for numeric data could use distance based,
        for categorical, we could use most frequent items.

        :param data: contain missing field data
        :param use_al_to_im: Whether or not to use algorithm to impute data
        :return: fitted estimator
        """
        # before we do any thing, should remove too many mssing cols
        self._remove_cols_up_to_threshold(data)

        # Let's first try to split data into 2 parts with numeric and category
        numeric_index, category_index = self.get_col_data_type(data)
        # we have to save it for later transformation use
        self.numeric_index = numeric_index
        self.category_index = category_index

        # get different data, there maybe without numeric or category index
        if self.numeric_index:
            numeric_data = data[:, self.numeric_index]
        if self.category_index:
            category_data = data[:, self.category_index]

        try:
            if self.use_al_to_im:
                # for KNN, we could only use numerical data
                if category_index:
                    raise ValueError("For algorithm logic shouldn't contain category data")

                self.num_estimator = KNNImputer(n_neighbors=5)
            else:
                # Here just use univariate processing logic for missing values
                if self.numeric_index:
                    self.num_estimator = SimpleImputer()
                if self.category_index:
                    self.cate_estimator = SimpleImputer(missing_values=np.nan, strategy='most_frequent')

            # we have to check different data type to process.
            if self.numeric_index:
                self.num_estimator.fit(numeric_data)

            if self.category_index:
                self.cate_estimator.fit(category_data.astype(np.object))

        except Exception as e:
            raise Exception("When try to impute data with Imputer with error: {}".format(e))

    def transform(self, data, y=None):
        # # before we do anything, first should keep the data that we want based on training data
        # data = self._transform_data_without_col_over_threshold(data)

        # here just to according to different type data to transform data and combine them
        if self.numeric_index:
            numeric_data = self.num_estimator.transform(data[:, self.numeric_index])

        if self.category_index:
            # we have to add `.astype(np.object)`, as imputation only work with object type
            cate_data = self.cate_estimator.transform(data[:, self.category_index].astype(np.object))

        # we could just drop the order of the data, then just combine them here
        # could do better with origin order
        if self.numeric_index and self.category_index:
            return np.concatenate([numeric_data, cate_data], axis=1)
        elif self.numeric_index:
            return numeric_data
        else:
            return cate_data

    def fit_transform(self, data, y=None):
        """
        This have to overwrite parent func, as we need different processing logic
        :param data:
        :return:
        """
        # first we have to do fit logic, just like sklearn done.
        self.fit(data, y=None)

        return self.transform(data, y=None)

    def _transform_data_without_col_over_threshold(self, data):
        """
        To keep the cols that are kept in the training process.
        :param data:
        :return:
        """
        if isinstance(data, pd.DataFrame):
            data = data[self.keep_cols]
        else:
            data = data[:, self.keep_cols]

        return data

    def _remove_cols_up_to_threshold(self, data):
        """
        As if there are too many missing value, we don't need it, just remove it

        Have to store the column that we want to keep for `transform` use case to remove
        the column that is over `threshold` NAN values.
        :param threshold: default is 0.5
        :return:
        """
        if isinstance(data, pd.DataFrame):
            null_ratio_series = data.isnull().sum() / len(data)
            self.keep_cols = list(null_ratio_series[null_ratio_series < self.threshold].index)
            data = data[self.keep_cols]
        else:
            null_ratio_series = pd.isnull(data).sum(axis=0) / len(data)
            self.keep_cols = np.array(range(data.shape[1]))[null_ratio_series < self.threshold]
            data = data[:, self.keep_cols]

        return data

    @staticmethod
    def get_col_data_type(data):
        """
        To get each column data type with some missing values
        :param data:
        :return:
        """
        # with two different types of feature: `numeric` and `category`
        if len(data.shape) == 1:
            data = data.reshape(-1, 1)
        n_cols = data.shape[1]
        numeric_col_index = []
        cate_col_index = []

        # just define a func to check value is nan or not
        def _is_nan(item):
            return item != item

        for i in range(n_cols):
            sample = data[:, i]
            # currently logic to try to get each value
            # todo: Must be changed, if there is column first as number, the other with string will get error.
            try:
                if all([isinstance(int(s), int) for s in sample if not _is_nan(s)]):
                    # if and only if whole value are number and the value is not NAN
                    numeric_col_index.append(i)
            except:
                cate_col_index.append(i)

            # while True:
            #     for j in range(len(sample)):
            #         item = sample[j]
            #         if _is_nan(item):
            #             # if face nan value, go to next
            #             continue
            #         else:
            #             # try to convert this data
            #             try:
            #                 if isinstance(int(item), int):
            #                     numeric_col_index.append(i)
            #             except:
            #                 cate_col_index.append(i)
            #             break
            #     break

        return numeric_col_index, cate_col_index

    # @staticmethod
    # def get_col_data_type(data):
    #     # with two different types of feature: `numeric` and `category`
    #     if len(data.shape) == 1:
    #         data = data.reshape(-1, 1)
    #     n_cols = data.shape[1]
    #     numeric_col_index = []
    #     cate_col_index = []
    #
    #     for i in range(data.shape[1]):
    #         if is_categorical_type(data[:, i]):
    #             cate_col_index.append(i)
    #         else:
    #             numeric_col_index.append(i)
    #
    #     return numeric_col_index, cate_col_index


if __name__ == '__main__':
    import numpy as np
    from sklearn.datasets import load_iris

    x, y = load_iris(return_X_y=True)

    sample_data = np.array([["1", 2], ["3", 6], ["4", 8], [np.nan, 3], ["4", np.nan]])
    data_new = np.array([['good', 1], [np.nan, 2]])
    lable = [1, 2]
    i = Imputation()
    print(i.get_col_data_type(x))
    # n, c = i.get_col_data_type(x)
    # print(x[:, n])
    # print(i.fit_transform(data_new))

    i.fit(data_new, lable)
    print(i.transform(data_new))
    print(i.fit_transform(sample_data))
    print(i.name)

    print("*" * 30)
    from auto_ml.test.get_test_data import get_training_data, save_processing_data

    x, y = get_training_data()
    i.fit(x, y)
    x_new = i.transform(x)

    save_processing_data(x_new, 'imputation')



# -*- coding:utf-8 -*-
"""
To do ensemble logic with whole trained model, try to improve whole score based on
different processing logic, also if we could get better result, then we are lucky!

One more thing, this should be called only and after the pipeline has finished, so that
we could load the trained model from disk, so this should be called from the parent
automl training logic.

@author: Guangqiang.lu
"""
import numpy as np
from sklearn.ensemble import VotingClassifier, VotingRegressor
from sklearn.model_selection import cross_validate
from auto_ml.utils.backend_obj import Backend
from auto_ml.metrics.scorer import accuracy, r2
from auto_ml.base.classifier_algorithms import ClassifierClass, ClassifierFactory
from auto_ml.utils.logger import create_logger
from auto_ml.utils.paths import load_yaml_file


logger = create_logger(__file__)


class ModelEnsemble(ClassifierClass):
    """
    Currently I want to support 2 different ensemble logic:
    Voting(weight combine classification: with soft voting and hard voting,
    regression: weight multiple)
    stacking(add trained model prediction into training data)
    """
    def __init__(self, backend, task_type='classification', ensemble_alg='voting',
                 voting_logic='soft'):
        """
        Based on different task to do different logic.
        :param task_type: which task to do: classification or regression.
        :param ensemble_alg: which ensemble logic to use: `voting` or `stacking`.
        :param voting_logic: whether with `hard` or `soft` voting
        """
        super().__init__()
        self.task_type = task_type
        self.ensemble_alg = ensemble_alg
        self.voting_logic = voting_logic
        # To load and save models
        # self.backend = backend if backend is not None else Backend()
        if backend is None:
            raise ValueError("When to use Model Ensemble class, we get a None `backend` object! Please check!")
        self.backend = backend

        self.model_list = self._load_trained_models()
        # define matrics based on task
        self.metrics = None
        if self.task_type == 'classification':
            self.metrics = accuracy
        elif self.task_type == 'regression':
            self.metrics = r2
        self.estimator = None
        # Whole trained model object list, like: [('LogisticRegression', LogisticRegression-9323.pkl),...]
        self.model_list_without_score = self._get_model_list_without_score()
        # Add attr for `stacking` logic to store the whole models needed to be load
        # for later step to make new dataset, so here just store the instances don't need to re-load
        self.stacking_models = [model[1] for model in self.model_list_without_score] \
            if ensemble_alg == 'stacking' else None

    def fit(self, x, y, **kwargs):
        # First we should to get whole trained models no matter for `voting` or `stacking`
        # self.model_list_without_score = self._get_model_list_without_score()

        if self.ensemble_alg == 'voting':
            self.fit_bagging(x, y, **kwargs)
        elif self.ensemble_alg == 'stacking':
            self.fit_stacking(x, y, **kwargs)

    def fit_bagging(self, x, y, **kwargs):
        """
        Here with ensemble logic like `hard` by number voting or
        `soft` by weight combine.

        For bagging fitting, if we face with classification problem,
        then we could use `voting` logic to get ensemble prediction,
        if regression, then will get weights * each model prediction.
        """
        # Here change logic with voting
        if self.task_type == 'classification':
            if self.voting_logic not in ['hard', 'soft']:
                raise ValueError("For ensemble logic, only `hard` and soft is supported "
                                 "when use `voting` logic.")

            self.estimator = VotingClassifier(estimators=self.model_list_without_score,
                                                voting=self.voting_logic)

            # start to fit the voting estimator
            self.estimator.fit(x, y)

            # Here add logic with saving trained model into disk.
            # TODO: but how to evaluate current model with score? Use cross-valiation to do score
            # get voting model score based on CV result!
            score = cross_validate(self.estimator, x, y, cv=3)['test_score'].mean()
            score_str = str(round(score, 6))
            logger.info("Model ensemble Cross-valiation score: {}".format(score_str))

            store_model_name = 'Voting_{}-{}'.format(self.voting_logic, score_str)

            logger.info("Start to save trained model: {} into disk.".format(store_model_name))
            self.backend.save_model(self.estimator, store_model_name)

        elif self.task_type == 'regression':
            pass

    def fit_stacking(self, x, y, **kwargs):
        """
        Implement with stacking logic is combined trained model prediction and original data into
        a new training data.

        # TODO: how TO choose new algorithm for the training?
        # Just to select the best score algorithm for the later step, based on the factory class
        to get the original class name and to load a new classifier.

        :param x: training data
        :param y: training label
        :param kwargs:
        :return:
        """
        # first should create new dataset.
        x_new = self.create_stacking_dataset(x, backend=self.backend)

        logger.info("Before stacking we have data dimention: {}, "
                    "after stacking we have :{}".format(x.shape[1], x_new.shape[1]))

        # Here we should get best score algorithm name for stacking
        # `model_list_without_score` is sorted based on score in fact.
        # In fact we have to add the estimator name based on what we have(using the yaml file result)
        # as we want to avoid to load the pipeline instance...
        best_estimator_name = self._get_best_model_estimator_name_based_on_yaml()

        logger.info("Get estaimator {} for stacking logic.".format(best_estimator_name))
        # As return is a list, but here we just need `one instance` for combined dataset
        self.estimator = ClassifierFactory.get_algorithm_instance(best_estimator_name)[0]

        logger.info("Loaded {} instance for `stacking` training.".format(best_estimator_name))
        # start training step for `stacking`
        self.estimator.fit(x_new, y)

        # model score should also based on CV result.
        score = cross_validate(self.estimator, x_new, y, cv=5)['test_score'][0]
        score = str(round(score, 6))
        stacking_model_name = "Stacking-{}".format(score)
        logger.info("Stacking model score: {}".format(score))

        self.backend.save_model(self.estimator, stacking_model_name)

    def _get_best_model_estimator_name_based_on_yaml(self):
        """
        Just to get the best estimator name based on the yaml file that we have
        for `stacking` ensemble logic.
        :return:
        """
        algorithm_name_list = load_yaml_file()['classification']['default']
        trained_model_alg_name_list = [x[0].split("-")[0] for x in self.model_list_without_score]

        for algo_name in trained_model_alg_name_list:
            if algo_name in algorithm_name_list:
                return algo_name

        return None

    def _load_trained_models(self):
        """
        To load whole trained model from disk, one more thing, as we also saved
        the processing model into disk, so here should ensure we just load the
        algorithm models.

        Sorted instance object list with name, also we could get the model score
        for later compare use case, this is `sorted` list, so later don't need to
        consider for the order.
        :return:a list of models: [('LR-0.98.pkl', LogisticRegression())]
        """
        model_list = self.backend.load_models_combined_with_model_name()

        # ADD logic: we shouldn't include the `Voting` algorithms instance object in fact
        model_list = [x for x in model_list if not x[0].lower().startswith('voting')]

        # Except Neural network models.
        model_list = [x for x in model_list if not x[0].endswith('.h5')]

        # To ensure there should be at least one file for `Ensemble` logic.
        if not model_list:
            raise IOError("There isn't any trained model for `Ensemble`.")

        # after we have get the model list, we should ensure the model by the model
        # name with endswith score.
        # Model name like this: ('lr_0.98.pkl', lr)
        # model_list.sort(key=lambda x: float("0." + x[0].split('-')[1].split('.')[0]), reverse=True)
        model_list.sort(key=lambda x: float(x[0].split('-')[1].split('.')[0]))

        return model_list

    def get_model_score_list(self):
        """
        To get each model accuracy score list for later compare
        :return:
        """
        score_list = []

        for model_name, _ in self._load_trained_models():
            model_score = model_name.split('.')[0].split('-')[-1]
            score_list.append(model_score)

        return score_list

    def _get_model_list_without_score(self):
        """
        To get the model list without score for ensemble use case.
        :return: [('LogisticRegression', LogisticRegression-9323.pkl),...]
        """
        model_list_without_score = []
        for estimator_tuple in self.model_list:
            estimator_name = estimator_tuple[0].split('-')[0]
            # we shouldn't include `ensemble` models.
            if not estimator_name.lower().startswith('votinig') \
                    and not estimator_name.lower().startswith('stacking'):
                model_list_without_score.append((estimator_name, estimator_tuple[1]))

        return model_list_without_score

    @classmethod
    def create_stacking_dataset(cls, x, backend, task_type='classification', ensemble_alg='stacking'):
        """
        What I want is to create the new dataset based on the whole instances for `stacking`.

        We could use the class func to create this.
        As `stacking` will add new features based on trained models.
        Should make the attr `stacking_models` with the `models instance`.
        :param x:
        :param task_type:
        :param ensemble_alg:
        :return:
        """
        model_ensemble = cls(backend=backend, task_type=task_type, ensemble_alg=ensemble_alg)

        # we don't need to care about `model_list_without_score` has instance or not, as parent does this.
        n_estimators = len(model_ensemble.model_list_without_score)

        # Whole trained estimator prediction result.
        pred_array = np.empty((len(x), n_estimators))

        logger.info("Start to get trained model prediction for `stacking`")
        model_list_without_score = model_ensemble.model_list_without_score
        for i in range(n_estimators):
            logger.info("To get prediction for estimator: {}".format(model_list_without_score[i][0]))

            estimator = model_list_without_score[i][1]

            pred = estimator.predict(x)
            pred_array[:, i] = pred

        # Then should combined the prediction and original data
        x_new = np.concatenate([x, pred_array], axis=1)

        return x_new


if __name__ == '__main__':
    from sklearn.datasets import load_iris
    from auto_ml.test.get_test_data import get_training_data

    x, y = load_iris(return_X_y=True)
    x, y = get_training_data()

    model_ensemble = ModelEnsemble(ensemble_alg='stacking', voting_logic='soft', )

    model_ensemble.fit(x, y)
    # print([x[1].__class__ for x in model_ensemble.model_list])
    # print(model_ensemble.model_list)
    # for models in model_ensemble.model_list:
    #     model = models[1]
    #     print(model)
    #     print(getattr(model, "_estimator_type", None))
    print(model_ensemble.stacking_models)

    print(ModelEnsemble.create_stacking_dataset(x))



# -*- coding:utf-8 -*-
"""
This is backend object that we could use for whole file
transfer or file processing, we could just use this
class to do like model saving, model loading etc.

Here also use `Singleton` method to create the object.

author: Guangqiang.lu
"""
import pickle
import shutil
from numpy.lib.arraysetops import isin
import pandas as pd
import traceback

try:
    is_tf_support = True
    from tensorflow.keras.models import load_model as keras_load_model
except ImportError:
    is_tf_support = False

from auto_ml.utils.logger import create_logger
from auto_ml.utils.CONSTANT import *
from auto_ml.utils.paths import load_yaml_file


logger = create_logger(__file__)

hyper_yml_file_name = 'search_hypers.yml'
# Load hyperparametersload_model
hyper_yml = load_yaml_file(hyper_yml_file_name)


class Backend(object):
    """
    This the the common module that could be used for whole project,
    currently supported with save models files, load model from disk,
    should be added more here.
    """
    # This is used for Singleton.
    instance = None

    def __init__(self,
                 tmp_folder_name=None,
                 output_folder=None,
                 delete_tmp_folder_after_terminate=True,
                 delete_output_folder_after_terminate=True):
        # This is to ensure we could use same folder for whole project.
        # but one more thing is if we face any problems, then we will lose whole
        self.tmp_folder_name = TMP_FOLDER if tmp_folder_name is None else tmp_folder_name
        self.output_folder = OUTPUT_FOLDER if output_folder is None else output_folder

        self.delete_tmp_folder_after_terminate = delete_tmp_folder_after_terminate
        self.delete_output_folder_after_terminate = delete_output_folder_after_terminate


        # I think when we init instance,then we should create the tmp folder first
        self._create_tmp_and_output_folder()

    def _create_tmp_and_output_folder(self):
        def __create_folder(folder_name):
            if not os.path.exists(folder_name):
                logger.info("Temp folder:{} don't exist, now try to create it.".format(folder_name))
                try:
                    os.makedirs(folder_name, exist_ok=True)
                except Exception as e:
                    logger.error("When create tmp and model folder with error: %s"
                                % traceback.format_exc())
                    raise IOError("When create tmp and model folder with error: %s" % e)
            else:
                logger.info("Folder: {} already exists.".format(folder_name))

        folder_list = [self.tmp_folder_name, self.output_folder]

        for folder in folder_list:
            __create_folder(folder)

    def delete_tmp_and_output_folder(self, delete_models=False):
        """As maybe we don't want to delete our trained model"""
        try:
            if self.delete_tmp_folder_after_terminate:
                shutil.rmtree(self.tmp_folder_name)

            if self.delete_output_folder_after_terminate and delete_models:
                shutil.rmtree(self.output_folder)
        except Exception as e:
            logger.exception("When try to remove temperate folder with error: %s" % e)

    def save_model(self, model, identifier):
        try:
            if not identifier.endswith(".pkl"):
                identifier += '.pkl'

            with open(os.path.join(self.output_folder, identifier), 'wb') as f:
                pickle.dump(model, f)

                logger.info("Model: {} has been saved into disk!".format(identifier))

        except Exception as e:
            logger.error("When try to save model: %s, face problem: %s" % (identifier, e))
            raise IOError("When try to save model: %s, face problem: %s" % (identifier, e))

    def load_model(self, identifier):
        """
        Get model instance object with just model name.
         Noted: `identifier` should contain with `extension`!

        :param identifier:
        :return:
        """
        # Here has to be changed for different platform
        default_neural_network_algorithms = hyper_yml['DefaultAlgorithms']
        identi_name = identifier.split('-')[0]
        if identi_name in default_neural_network_algorithms:
            # Keras model.
            return self.load_keras_model(identifier)

        # For sklearn models by using `pickle`
        try:
            model_file_path = os.path.join(self.output_folder, identifier)
            self._file_exists(model_file_path)

            with open(os.path.join(self.output_folder, identifier), 'rb') as f:
                model = pickle.load(f)

            return model
        except Exception as e:
            logger.error("When load %s model with error: %s." %
                         (identifier, traceback.format_exc()))
            raise IOError("When to load %s model, the file:%s not exist!" % (identifier, e))

    def load_keras_model(self, identifier):
        if not identifier.endswith('h5'):
            raise ValueError("To load keras model, model extension must end with `h5`")
        
        try:
            model_path = os.path.join(self.output_folder, identifier)
            
            self._file_exists(model_path)
            if not is_tf_support:
                raise RuntimeError("When try to use `TensorFlow` backend to load model, couldn't load module: `tensorflow`, please check!")

            model = keras_load_model(model_path)

            return model
        except IOError as e:
            logger.error("When to load keras model, get error: {}".format(e))
            raise IOError("When to load keras model, get error: {}".format(e))

    def _file_exists(self, file_path):
        """Check file exist or not based on path.

        Args:
            file_path (Str): File path to check.

        Raises:
            IOError: [description]
        """
        if not os.path.exists(file_path):
                raise IOError("File: {} not exist!".format(file_path))

    def load_models_by_identi_combined_with_model_name(self, identifiers):
        """
        As I don't want to change current logic with `load_model` to just
        return model instance, sometimes we also need the model name, so
        that we could also return (model_name, model_instance)
        :param identifiers:
        :return:
        """
        model_list = []

        if isinstance(identifiers, str):
            identifiers = [identifiers]

        for identi in identifiers:
            model_list.append((identi, self.load_model(identi)))

        return model_list

    def list_models(self, extension=['pkl', 'h5'], except_model_list=["processing_pipeline"]):
        """
        this is to list whole saved model in local disk,
        model name should be end with `pkl`, should be extended
        to other framework with other extension.
        :return: a list of model list
        """
        models_list = []

        # change this with keras models.
        if extension and isinstance(extension, list):
            for ex in extension:
                models_list.extend([x for x in os.listdir(self.output_folder) if x.endswith(ex)])
        # models_list = [x for x in os.listdir(self.output_folder) if x.endswith(extension)]
        
        if not models_list:
            logger.warning("There isn't any trained models in folder: {}".format(self.output_folder))
        

        # Here add logic to ensure we just add algorithm model list
        if except_model_list:
            models_list = [x for x in models_list if x.split('.')[0] not in except_model_list]

        return models_list

    def list_models_with_identifiers(self, identifiers, extension='pkl'):
        """
        this is to list whole models with identifiers satisfication.
        :param identifiers: a list of idenfifiers
        :param extension: what data type to include.
        :return: a model list with satisfied
        """
        models_list = [x for x in os.listdir(self.output_folder)
                       if x.endswith(extension) and x.split('.')[0] in identifiers]

        return models_list

    def load_models(self):
        """
        this is to load whole models already trained in disk.
        This should be algorithm models.
        :return: a list of trained model object.
        """
        model_obj_list = []
        for identifier in self.list_models():
            model_obj_list.append(self.load_model(identifier))
        return model_obj_list

    def load_models_combined_with_model_name(self):
        """
        To load models combined with `model name`, so later could use this
        models based on model score.
        :return: a list of models: [('LR-0.98.pkl', LogisticRegression-0.982323.pkl)]
        """
        model_with_name_list = []

        for identifier in self.list_models():
            model_with_name_list.extend(
                self.load_models_by_identi_combined_with_model_name(identifier))

        return model_with_name_list

    def load_models_with_identifiers(self, identifiers):
        """
        this is to load models with identifiers
        :param identifiers: a list of identifiers
        :return: a list of trained model object
        """
        model_obj_list = []
        for identifier in identifiers:
            model_obj_list.append(self.load_model(identifier))
        return model_obj_list

    def save_dataset(self, dataset, dataset_name, model_file_path=True):
        """
        As most of us have the structed data, so that I would love to
        store the data into disk with `csv` file, then we could use
        `pandas` to load it from the disk and convert it directly into
        DataFrame.
        :param dataset: dataset object to be saved
        :param dataset_name: what name for it, so that we could re-load it
        :param file_path: Where to save the data.
        :return:
        """
        if model_file_path:
            file_path = self.output_folder
        else:
            file_path = self.tmp_folder_name

        try:
            dataset_path = os.path.join(file_path, dataset_name + '.csv')
            if not isinstance(dataset, pd.DataFrame):
                dataset = pd.DataFrame(dataset, columns=[str(x) for x in range(1, dataset.shape[1] + 1)])

            dataset.to_csv(dataset_path, index=False, sep='|')
            logger.info("Dataset: {} has been saved into: {}".format(dataset_name, dataset_path))
        except IOError as e:
            raise IOError("When try to save dataset: {} get error: {}".format(dataset_name, e))

    def load_dataset(self, dataset_name, file_path=None, sep="|"):
        """
        To load csv data again from disk.
        :param dataset_name:
        :param file_path:
        :return:
        """
        if file_path is None:
            file_path = self.tmp_folder_name

        # we should ensure the dataset is endswith 'csv'
        if not dataset_name.endswith(".csv"):
            dataset_name += '.csv'

        if dataset_name not in os.listdir(file_path):
            raise IOError("When try to load dataset: {} from path:{}, "
                          "we couldn't find it!".format(dataset_name, file_path))

        try:
            dataset_path = os.path.join(file_path, dataset_name)
            dataset = pd.read_csv(dataset_path, sep=sep)

            return dataset
        except IOError as e:
            raise IOError("When try to load dataset: {} get error: {}".format(dataset_name, e))

    def clean_folder(self, folder_name=None):
        """
        To clean `folder_name` to ensure every time we could get a empty folder.
        :param folder_name: default is `models`
        :return:
        """
        if folder_name is None:
            # default is model path
            folder_name = self.output_folder
        
        if os.path.exists(folder_name):
            file_list = os.listdir(folder_name)
            if file_list:
                for file in file_list:
                    try:
                        os.remove(os.path.join(folder_name, file))
                    except IOError as e:
                        # in case if need with many trials couldn't delete these files, not good idea, so raise will be a solution
                        raise IOError("When try to remove file: {} in folder:{} get error: {}".format(file, folder_name, e))
        
    @classmethod
    def __new__(cls, *args, **kwargs):
        """
        `Singleton` is used to ensure there will be just one instance in
        whole running process.
        :param args:
        :param kwargs:
        :return:
        """
        if not cls.instance:
            # we should avoid to use __new__(cls, *args, **kwargs) as with error need to fix, just with __new__(cls) will be fine.
            # https://stackoverflow.com/questions/34777773/typeerror-object-takes-no-parameters-after-defining-new
            cls.instance = super(Backend, cls).__new__(cls)
            
        return cls.instance


if __name__ == "__main__":
    backend = Backend()
    print(backend.output_folder)
    print(backend.tmp_folder_name)

    # print(backend.load_models_combined_with_model_name())

    # # To test there is just one instance.
    # print(backend == Backend())
